#!/usr/bin/python

# Copyright (c) 2016, Shinichi TAMURA (@tmshn)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: timezone
short_description: Configure timezone setting
description:
  - This module configures the timezone setting, both of the system clock and of the hardware clock. If you want to set up
    the NTP, use M(ansible.builtin.service) module.
  - It is recommended to restart C(crond) after changing the timezone, otherwise the jobs may run at the wrong time.
  - Several different tools are used depending on the OS/Distribution involved. For Linux it can use C(timedatectl) or edit
    C(/etc/sysconfig/clock) or C(/etc/timezone) and C(hwclock). On SmartOS, C(sm-set-timezone), for macOS, C(systemsetup),
    for BSD, C(/etc/localtime) is modified. On AIX, C(chtz) is used.
  - Make sure that the zoneinfo files are installed with the appropriate OS package, like C(tzdata) (usually always installed,
    when not using a minimal installation like Alpine Linux).
  - Windows and HPUX are not supported, please let us know if you find any other OS/distro in which this fails.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
options:
  name:
    description:
      - Name of the timezone for the system clock.
      - Default is to keep current setting.
      - B(At least one) of O(name) and O(hwclock) are required.
    type: str
  hwclock:
    description:
      - Whether the hardware clock is in UTC or in local timezone.
      - Default is to keep current setting.
      - Note that this option is recommended not to change and may fail to configure, especially on virtual environments such
        as AWS.
      - B(At least one) of O(name) and O(hwclock) are required.
      - I(Only used on Linux).
    type: str
    aliases: [rtc]
    choices: [local, UTC]
notes:
  - On Ubuntu 24.04 and Debian 13 (Trixie), the C(util-linux-extra) package is required to provide the C(hwclock) command.
  - On SmartOS the C(sm-set-timezone) utility (part of the smtools package) is required to set the zone timezone.
  - On AIX only Olson/tz database timezones are usable (POSIX is not supported). An OS reboot is also required on AIX for
    the new timezone setting to take effect. Note that AIX 6.1+ is needed (OS level 61 or newer).
author:
  - Shinichi TAMURA (@tmshn)
  - Jasper Lievisse Adriaanse (@jasperla)
  - Indrajit Raychaudhuri (@indrajitr)
"""

EXAMPLES = r"""
- name: Set timezone to Asia/Tokyo
  become: true
  community.general.timezone:
    name: Asia/Tokyo

- name: Set timezone and hardware clock to UTC
  community.general.timezone:
    name: UTC
    hwclock: UTC
"""

import errno
import filecmp
import os
import platform
import random
import re
import string

from ansible.module_utils.basic import AnsibleModule, get_distribution


class Timezone:
    """This is a generic Timezone manipulation class that is subclassed based on platform.

    A subclass may wish to override the following action methods:
        - get(key, phase)   ... get the value from the system at `phase`
        - set(key, value)   ... set the value to the current system
    """

    def __new__(cls, module):
        """Return the platform-specific subclass.

        It does not use load_platform_subclass() because it needs to judge based
        on whether the `timedatectl` command exists and is available.

        Args:
            module: The AnsibleModule.
        """
        if platform.system() == "Linux":
            timedatectl = module.get_bin_path("timedatectl")
            if timedatectl is not None:
                rc, stdout, stderr = module.run_command(timedatectl)
                if rc == 0:
                    return super(Timezone, SystemdTimezone).__new__(SystemdTimezone)
                else:
                    module.debug(f"timedatectl command was found but not usable: {stderr}. using other method.")
                    return super(Timezone, NosystemdTimezone).__new__(NosystemdTimezone)
            else:
                return super(Timezone, NosystemdTimezone).__new__(NosystemdTimezone)
        elif re.match("^joyent_.*Z", platform.version()):
            # platform.system() returns SunOS, which is too broad. So look at the
            # platform version instead. However we have to ensure that we're not
            # running in the global zone where changing the timezone has no effect.
            zonename_cmd = module.get_bin_path("zonename")
            if zonename_cmd is not None:
                (rc, stdout, dummy) = module.run_command(zonename_cmd)
                if rc == 0 and stdout.strip() == "global":
                    module.fail_json(msg="Adjusting timezone is not supported in Global Zone")

            return super(Timezone, SmartOSTimezone).__new__(SmartOSTimezone)
        elif platform.system() == "Darwin":
            return super(Timezone, DarwinTimezone).__new__(DarwinTimezone)
        elif re.match("^(Free|Net|Open)BSD", platform.platform()):
            return super(Timezone, BSDTimezone).__new__(BSDTimezone)
        elif platform.system() == "AIX":
            AIXoslevel = int(platform.version() + platform.release())
            if AIXoslevel >= 61:
                return super(Timezone, AIXTimezone).__new__(AIXTimezone)
            else:
                module.fail_json(msg=f"AIX os level must be >= 61 for timezone module (Target: {AIXoslevel}).")
        else:
            # Not supported yet
            return super(Timezone, Timezone).__new__(Timezone)

    def __init__(self, module):
        """Initialize of the class.

        Args:
            module: The AnsibleModule.
        """
        super().__init__()
        self.msg = []
        # `self.value` holds the values for each params on each phases.
        # Initially there's only info of "planned" phase, but the
        # `self.check()` function will fill out it.
        self.value = dict()
        for key in module.argument_spec:
            value = module.params[key]
            if value is not None:
                self.value[key] = dict(planned=value)
        self.module = module

    def abort(self, msg):
        """Abort the process with error message.

        This is just the wrapper of module.fail_json().

        Args:
            msg: The error message.
        """
        error_msg = ["Error message:", msg]
        if len(self.msg) > 0:
            error_msg.append("Other message(s):")
            error_msg.extend(self.msg)
        self.module.fail_json(msg="\n".join(error_msg))

    def execute(self, *commands, **kwargs):
        """Execute the shell command.

        This is just the wrapper of module.run_command().

        Args:
            *commands: The command to execute.
            **kwargs:  Only 'log' key is checked.
                If kwargs['log'] is true, record the command to self.msg.

        Returns:
            stdout: Standard output of the command.
        """
        (rc, stdout, stderr) = self.module.run_command(list(commands), check_rc=True)
        if kwargs.get("log", False):
            self.msg.append(f"executed `{' '.join(commands)}`")
        return stdout

    def diff(self, phase1="before", phase2="after"):
        """Calculate the difference between given 2 phases.

        Args:
            phase1, phase2: The names of phase to compare.

        Returns:
            diff: The difference of value between phase1 and phase2.
                This is in the format which can be used with the
                `--diff` option of ansible-playbook.
        """
        diff = {phase1: {}, phase2: {}}
        for key, value in self.value.items():
            diff[phase1][key] = value[phase1]
            diff[phase2][key] = value[phase2]
        return diff

    def check(self, phase):
        """Check the state in given phase and set it to `self.value`.

        Args:
            phase: The name of the phase to check.

        Returns:
            NO RETURN VALUE
        """
        if phase == "planned":
            return
        for key, value in self.value.items():
            value[phase] = self.get(key, phase)

    def change(self):
        """Make the changes effect based on `self.value`."""
        for key, value in self.value.items():
            if value["before"] != value["planned"]:
                self.set(key, value["planned"])

    # ===========================================
    # Platform specific methods (must be replaced by subclass).

    def get(self, key, phase):
        """Get the value for the key at the given phase.

        Called from self.check().

        Args:
            key:   The key to get the value
            phase: The phase to get the value

        Return:
            value: The value for the key at the given phase.
        """
        self.abort("get(key, phase) is not implemented on target platform")

    def set(self, key, value):
        """Set the value for the key (of course, for the phase 'after').

        Called from self.change().

        Args:
            key: Key to set the value
            value: Value to set
        """
        self.abort("set(key, value) is not implemented on target platform")

    def _verify_timezone(self):
        tz = self.value["name"]["planned"]
        tzfile = f"/usr/share/zoneinfo/{tz}"
        if not os.path.isfile(tzfile):
            self.abort(f'given timezone "{tz}" is not available')
        return tzfile


class SystemdTimezone(Timezone):
    """This is a Timezone manipulation class for systemd-powered Linux.

    It uses the `timedatectl` command to check/set all arguments.
    """

    regexps = dict(
        hwclock=re.compile(r"^\s*RTC in local TZ\s*:\s*([^\s]+)", re.MULTILINE),
        name=re.compile(r"^\s*Time ?zone\s*:\s*([^\s]+)", re.MULTILINE),
    )

    subcmds = dict(hwclock="set-local-rtc", name="set-timezone")

    def __init__(self, module):
        super().__init__(module)
        self.timedatectl = module.get_bin_path("timedatectl", required=True)
        self.status = dict()
        # Validate given timezone
        if "name" in self.value:
            self._verify_timezone()

    def _get_status(self, phase):
        if phase not in self.status:
            self.status[phase] = self.execute(self.timedatectl, "status")
        return self.status[phase]

    def get(self, key, phase):
        status = self._get_status(phase)
        value = self.regexps[key].search(status).group(1)
        if key == "hwclock":
            # For key='hwclock'; convert yes/no -> local/UTC
            if self.module.boolean(value):
                value = "local"
            else:
                value = "UTC"
        return value

    def set(self, key, value):
        # For key='hwclock'; convert UTC/local -> yes/no
        if key == "hwclock":
            if value == "local":
                value = "yes"
            else:
                value = "no"
        self.execute(self.timedatectl, self.subcmds[key], value, log=True)


class NosystemdTimezone(Timezone):
    """This is a Timezone manipulation class for non systemd-powered Linux.

    For timezone setting, it edits the following file and reflect changes:
        - /etc/sysconfig/clock  ... RHEL/CentOS
        - /etc/timezone         ... Debian/Ubuntu
    For hwclock setting, it executes `hwclock --systohc` command with the
    '--utc' or '--localtime' option.
    """

    conf_files = dict(
        name=None,  # To be set in __init__
        hwclock=None,  # To be set in __init__
        adjtime="/etc/adjtime",
    )

    # It is fine if all tree config files don't exist
    allow_no_file = dict(name=True, hwclock=True, adjtime=True)

    regexps = dict(
        name=None,  # To be set in __init__
        hwclock=re.compile(r"^UTC\s*=\s*([^\s]+)", re.MULTILINE),
        adjtime=re.compile(r"^(UTC|LOCAL)$", re.MULTILINE),
    )

    dist_regexps = dict(
        SuSE=re.compile(r'^TIMEZONE\s*=\s*"?([^"\s]+)"?', re.MULTILINE),
        redhat=re.compile(r'^ZONE\s*=\s*"?([^"\s]+)"?', re.MULTILINE),
    )

    dist_tzline_format = dict(SuSE='TIMEZONE="%s"\n', redhat='ZONE="%s"\n')

    def __init__(self, module):
        super().__init__(module)
        # Validate given timezone
        planned_tz = ""
        if "name" in self.value:
            tzfile = self._verify_timezone()
            planned_tz = self.value["name"]["planned"]
            # `--remove-destination` is needed if /etc/localtime is a symlink so
            # that it overwrites it instead of following it.
            self.update_timezone = [
                [self.module.get_bin_path("cp", required=True), "--remove-destination", tzfile, "/etc/localtime"]
            ]
        self.update_hwclock = self.module.get_bin_path("hwclock", required=True)
        distribution = get_distribution()
        self.conf_files["name"] = "/etc/timezone"
        self.regexps["name"] = re.compile(r"^([^\s]+)", re.MULTILINE)
        self.tzline_format = "%s\n"
        # Distribution-specific configurations
        if self.module.get_bin_path("dpkg-reconfigure") is not None:
            # Debian/Ubuntu
            if "name" in self.value:
                self.update_timezone = [
                    [self.module.get_bin_path("ln", required=True), "-sf", tzfile, "/etc/localtime"],
                    [
                        self.module.get_bin_path("dpkg-reconfigure", required=True),
                        "--frontend",
                        "noninteractive",
                        "tzdata",
                    ],
                ]
            self.conf_files["hwclock"] = "/etc/default/rcS"
        elif distribution == "Alpine" or distribution == "Gentoo":
            self.conf_files["hwclock"] = "/etc/conf.d/hwclock"
            if distribution == "Alpine":
                self.update_timezone = [[self.module.get_bin_path("setup-timezone", required=True), "-z", planned_tz]]
        else:
            # RHEL/CentOS/SUSE
            if self.module.get_bin_path("tzdata-update") is not None:
                # tzdata-update cannot update the timezone if /etc/localtime is
                # a symlink so we have to use cp to update the time zone which
                # was set above.
                if not os.path.islink("/etc/localtime"):
                    self.update_timezone = [[self.module.get_bin_path("tzdata-update", required=True)]]
                # else:
                #   self.update_timezone       = 'cp --remove-destination ...' <- configured above
            self.conf_files["name"] = "/etc/sysconfig/clock"
            self.conf_files["hwclock"] = "/etc/sysconfig/clock"
            try:
                with open(self.conf_files["name"]) as f:
                    sysconfig_clock = f.read()
            except OSError as err:
                if self._allow_ioerror(err, "name"):
                    # If the config file doesn't exist detect the distribution and set regexps.
                    if distribution == "SuSE":
                        # For SUSE
                        self.regexps["name"] = self.dist_regexps["SuSE"]
                        self.tzline_format = self.dist_tzline_format["SuSE"]
                    else:
                        # For RHEL/CentOS
                        self.regexps["name"] = self.dist_regexps["redhat"]
                        self.tzline_format = self.dist_tzline_format["redhat"]
                else:
                    self.abort(f'could not read configuration file "{self.conf_files["name"]}"')
            else:
                # The key for timezone might be `ZONE` or `TIMEZONE`
                # (the former is used in RHEL/CentOS and the latter is used in SUSE linux).
                # So check the content of /etc/sysconfig/clock and decide which key to use.
                if re.search(r"^TIMEZONE\s*=", sysconfig_clock, re.MULTILINE):
                    # For SUSE
                    self.regexps["name"] = self.dist_regexps["SuSE"]
                    self.tzline_format = self.dist_tzline_format["SuSE"]
                else:
                    # For RHEL/CentOS
                    self.regexps["name"] = self.dist_regexps["redhat"]
                    self.tzline_format = self.dist_tzline_format["redhat"]

    def _allow_ioerror(self, err, key):
        # In some cases, even if the target file does not exist,
        # simply creating it may solve the problem.
        # In such cases, we should continue the configuration rather than aborting.
        if err.errno != errno.ENOENT:
            # If the error is not ENOENT ("No such file or directory"),
            # (e.g., permission error, etc), we should abort.
            return False
        return self.allow_no_file.get(key, False)

    def _edit_file(self, filename, regexp, value, key):
        """Replace the first matched line with given `value`.

        If `regexp` matched more than once, other than the first line will be deleted.

        Args:
            filename: The name of the file to edit.
            regexp:   The regular expression to search with.
            value:    The line which will be inserted.
            key:      For what key the file is being edited.
        """
        # Read the file
        try:
            with open(filename) as file:
                lines = file.readlines()
        except OSError as err:
            if self._allow_ioerror(err, key):
                lines = []
            else:
                self.abort(f'tried to configure {key} using a file "{filename}", but could not read it')
        # Find the all matched lines
        matched_indices = []
        for i, line in enumerate(lines):
            if regexp.search(line):
                matched_indices.append(i)
        if len(matched_indices) > 0:
            insert_line = matched_indices[0]
        else:
            insert_line = 0
        # Remove all matched lines
        for i in matched_indices[::-1]:
            del lines[i]
        # ...and insert the value
        lines.insert(insert_line, value)
        # Write the changes
        try:
            with open(filename, "w") as file:
                file.writelines(lines)
        except OSError:
            self.abort(f'tried to configure {key} using a file "{filename}", but could not write to it')
        self.msg.append(f"Added 1 line and deleted {len(matched_indices)} line(s) on {filename}")

    def _get_value_from_config(self, key, phase):
        filename = self.conf_files[key]
        try:
            with open(filename) as file:
                status = file.read()
        except OSError as err:
            if self._allow_ioerror(err, key):
                if key == "hwclock":
                    return "n/a"
                elif key == "adjtime":
                    return "UTC"
                elif key == "name":
                    return "n/a"
            else:
                self.abort(f'tried to configure {key} using a file "{filename}", but could not read it')
        else:
            try:
                value = self.regexps[key].search(status).group(1)
            except AttributeError:
                if key == "hwclock":
                    # If we cannot find UTC in the config that's fine.
                    return "n/a"
                elif key == "adjtime":
                    # If we cannot find UTC/LOCAL in /etc/cannot that means UTC
                    # will be used by default.
                    return "UTC"
                elif key == "name":
                    if phase == "before":
                        # In 'before' phase UTC/LOCAL doesn't need to be set in
                        # the timezone config file, so we ignore this error.
                        return "n/a"
                    else:
                        self.abort(
                            f'tried to configure {key} using a file "{filename}", but could not find a valid value in it'
                        )
            else:
                if key == "hwclock":
                    # convert yes/no -> UTC/local
                    if self.module.boolean(value):
                        value = "UTC"
                    else:
                        value = "local"
                elif key == "adjtime":
                    # convert LOCAL -> local
                    if value != "UTC":
                        value = value.lower()
        return value

    def get(self, key, phase):
        planned = self.value[key]["planned"]
        if key == "hwclock":
            value = self._get_value_from_config(key, phase)
            if value == planned:
                # If the value in the config file is the same as the 'planned'
                # value, we need to check /etc/adjtime.
                value = self._get_value_from_config("adjtime", phase)
        elif key == "name":
            value = self._get_value_from_config(key, phase)
            if value == planned:
                # If the planned values is the same as the one in the config file
                # we need to check if /etc/localtime is also set to the 'planned' zone.
                if os.path.islink("/etc/localtime"):
                    # If /etc/localtime is a symlink and is not set to the TZ we 'planned'
                    # to set, we need to return the TZ which the symlink points to.
                    if os.path.exists("/etc/localtime"):
                        # We use readlink() because on some distros zone files are symlinks
                        # to other zone files, so it is hard to get which TZ is actually set
                        # if we follow the symlink.
                        path = os.readlink("/etc/localtime")
                        # most linuxes has it in /usr/share/zoneinfo
                        # alpine linux links under /etc/zoneinfo
                        linktz = re.search(r"(?:/(?:usr/share|etc)/zoneinfo/)(.*)", path, re.MULTILINE)
                        if linktz:
                            valuelink = linktz.group(1)
                            if valuelink != planned:
                                value = valuelink
                        else:
                            # Set current TZ to 'n/a' if the symlink points to a path
                            # which isn't a zone file.
                            value = "n/a"
                    else:
                        # Set current TZ to 'n/a' if the symlink to the zone file is broken.
                        value = "n/a"
                else:
                    # If /etc/localtime is not a symlink best we can do is compare it with
                    # the 'planned' zone info file and return 'n/a' if they are different.
                    try:
                        if not filecmp.cmp("/etc/localtime", f"/usr/share/zoneinfo/{planned}"):
                            return "n/a"
                    except Exception:
                        return "n/a"
        else:
            self.abort(f'unknown parameter "{key}"')
        return value

    def set_timezone(self, value):
        self._edit_file(
            filename=self.conf_files["name"], regexp=self.regexps["name"], value=self.tzline_format % value, key="name"
        )
        for cmd in self.update_timezone:
            self.execute(*cmd)

    def set_hwclock(self, value):
        if value == "local":
            option = "--localtime"
            utc = "no"
        else:
            option = "--utc"
            utc = "yes"
        if self.conf_files["hwclock"] is not None:
            self._edit_file(
                filename=self.conf_files["hwclock"], regexp=self.regexps["hwclock"], value=f"UTC={utc}\n", key="hwclock"
            )
        self.execute(self.update_hwclock, "--systohc", option, log=True)

    def set(self, key, value):
        if key == "name":
            self.set_timezone(value)
        elif key == "hwclock":
            self.set_hwclock(value)
        else:
            self.abort(f'unknown parameter "{key}"')


class SmartOSTimezone(Timezone):
    """This is a Timezone manipulation class for SmartOS instances.

    It uses the C(sm-set-timezone) utility to set the timezone, and
    inspects C(/etc/default/init) to determine the current timezone.

    NB: A zone needs to be rebooted in order for the change to be
    activated.
    """

    def __init__(self, module):
        super().__init__(module)
        self.settimezone = self.module.get_bin_path("sm-set-timezone", required=False)
        if not self.settimezone:
            module.fail_json(msg="sm-set-timezone not found. Make sure the smtools package is installed.")

    def get(self, key, phase):
        """Lookup the current timezone name in `/etc/default/init`. If anything else
        is requested, or if the TZ field is not set we fail.
        """
        if key == "name":
            try:
                with open("/etc/default/init") as f:
                    for line in f:
                        m = re.match("^TZ=(.*)$", line.strip())
                        if m:
                            return m.groups()[0]
            except Exception:
                self.module.fail_json(msg="Failed to read /etc/default/init")
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")

    def set(self, key, value):
        """Set the requested timezone through sm-set-timezone, an invalid timezone name
        will be rejected and we have no further input validation to perform.
        """
        if key == "name":
            cmd = ["sm-set-timezone", value]

            (rc, stdout, stderr) = self.module.run_command(cmd)

            if rc != 0:
                self.module.fail_json(msg=stderr)

            # sm-set-timezone knows no state and will always set the timezone.
            # XXX: https://github.com/joyent/smtools/pull/2
            m = re.match(rf"^\* Changed (to)? timezone (to)? ({value}).*", stdout.splitlines()[1])
            if not (m and m.groups()[-1] == value):
                self.module.fail_json(msg="Failed to set timezone")
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")


class DarwinTimezone(Timezone):
    """This is the timezone implementation for Darwin which, unlike other *BSD
    implementations, uses the `systemsetup` command on Darwin to check/set
    the timezone.
    """

    regexps = dict(name=re.compile(r"^\s*Time ?Zone\s*:\s*([^\s]+)", re.MULTILINE))

    def __init__(self, module):
        super().__init__(module)
        self.systemsetup = module.get_bin_path("systemsetup", required=True)
        self.status = dict()
        # Validate given timezone
        if "name" in self.value:
            self._verify_timezone()

    def _get_current_timezone(self, phase):
        """Lookup the current timezone via `systemsetup -gettimezone`."""
        if phase not in self.status:
            self.status[phase] = self.execute(self.systemsetup, "-gettimezone")
        return self.status[phase]

    def _verify_timezone(self):
        tz = self.value["name"]["planned"]
        # Lookup the list of supported timezones via `systemsetup -listtimezones`.
        # Note: Skip the first line that contains the label 'Time Zones:'
        out = self.execute(self.systemsetup, "-listtimezones").splitlines()[1:]
        tz_list = [x.strip() for x in out]
        if tz not in tz_list:
            self.abort(f'given timezone "{tz}" is not available')
        return tz

    def get(self, key, phase):
        if key == "name":
            status = self._get_current_timezone(phase)
            value = self.regexps[key].search(status).group(1)
            return value
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")

    def set(self, key, value):
        if key == "name":
            self.execute(self.systemsetup, "-settimezone", value, log=True)
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")


class BSDTimezone(Timezone):
    """This is the timezone implementation for *BSD which works simply through
    updating the `/etc/localtime` symlink to point to a valid timezone name under
    `/usr/share/zoneinfo`.
    """

    def __init__(self, module):
        super().__init__(module)

    def __get_timezone(self):
        zoneinfo_dir = "/usr/share/zoneinfo/"
        localtime_file = "/etc/localtime"

        # Strategy 1:
        #   If /etc/localtime does not exist, assume the timezone is UTC.
        if not os.path.exists(localtime_file):
            self.module.warn("Could not read /etc/localtime. Assuming UTC.")
            return "UTC"

        # Strategy 2:
        #   Follow symlink of /etc/localtime
        zoneinfo_file = localtime_file
        while not zoneinfo_file.startswith(zoneinfo_dir):
            try:
                zoneinfo_file = os.readlink(localtime_file)
            except OSError:
                # OSError means "end of symlink chain" or broken link.
                break
        else:
            return zoneinfo_file.replace(zoneinfo_dir, "")

        # Strategy 3:
        #   (If /etc/localtime is not symlinked)
        #   Check all files in /usr/share/zoneinfo and return first non-link match.
        for dname, dummy, fnames in sorted(os.walk(zoneinfo_dir)):
            for fname in sorted(fnames):
                zoneinfo_file = os.path.join(dname, fname)
                if not os.path.islink(zoneinfo_file) and filecmp.cmp(zoneinfo_file, localtime_file):
                    return zoneinfo_file.replace(zoneinfo_dir, "")

        # Strategy 4:
        #   As a fall-back, return 'UTC' as default assumption.
        self.module.warn("Could not identify timezone name from /etc/localtime. Assuming UTC.")
        return "UTC"

    def get(self, key, phase):
        """Lookup the current timezone by resolving `/etc/localtime`."""
        if key == "name":
            return self.__get_timezone()
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")

    def set(self, key, value):
        if key == "name":
            # First determine if the requested timezone is valid by looking in
            # the zoneinfo directory.
            zonefile = f"/usr/share/zoneinfo/{value}"
            try:
                if not os.path.isfile(zonefile):
                    self.module.fail_json(msg=f"{value} is not a recognized timezone")
            except Exception:
                self.module.fail_json(msg=f"Failed to stat {zonefile}")

            # Now (somewhat) atomically update the symlink by creating a new
            # symlink and move it into place. Otherwise we have to remove the
            # original symlink and create the new symlink, however that would
            # create a race condition in case another process tries to read
            # /etc/localtime between removal and creation.
            suffix = "".join([random.choice(string.ascii_letters + string.digits) for x in range(0, 10)])
            new_localtime = f"/etc/localtime.{suffix}"

            try:
                os.symlink(zonefile, new_localtime)
                os.rename(new_localtime, "/etc/localtime")
            except Exception:
                os.remove(new_localtime)
                self.module.fail_json(msg="Could not update /etc/localtime")
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")


class AIXTimezone(Timezone):
    """This is a Timezone manipulation class for AIX instances.

    It uses the C(chtz) utility to set the timezone, and
    inspects C(/etc/environment) to determine the current timezone.

    While AIX time zones can be set using two formats (POSIX and
    Olson) the preferred method is Olson.
    See the following article for more information:
    https://developer.ibm.com/articles/au-aix-posix/

    NB: AIX needs to be rebooted in order for the change to be
    activated.
    """

    def __init__(self, module):
        super().__init__(module)
        self.settimezone = self.module.get_bin_path("chtz", required=True)

    def __get_timezone(self):
        """Return the current value of TZ= in /etc/environment"""
        try:
            with open("/etc/environment") as f:
                etcenvironment = f.read()
        except Exception:
            self.module.fail_json(msg="Issue reading contents of /etc/environment")

        match = re.search(r"^TZ=(.*)$", etcenvironment, re.MULTILINE)
        if match:
            return match.group(1)
        else:
            return None

    def get(self, key, phase):
        """Lookup the current timezone name in `/etc/environment`. If anything else
        is requested, or if the TZ field is not set we fail.
        """
        if key == "name":
            return self.__get_timezone()
        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")

    def set(self, key, value):
        """Set the requested timezone through chtz, an invalid timezone name
        will be rejected and we have no further input validation to perform.
        """
        if key == "name":
            # chtz seems to always return 0 on AIX 7.2, even for invalid timezone values.
            # It will only return non-zero if the chtz command itself fails, it does not check for
            #  valid timezones. We need to perform a basic check to confirm that the timezone
            #  definition exists in /usr/share/lib/zoneinfo
            # This does mean that we can only support Olson for now. The below commented out regex
            #  detects Olson date formats, so in the future we could detect Posix or Olson and
            #  act accordingly.

            # regex_olson = re.compile('^([a-z0-9_\-\+]+\/?)+$', re.IGNORECASE)
            # if not regex_olson.match(value):
            #     msg = 'Supplied timezone (%s) does not appear to a be valid Olson string' % value
            #     self.module.fail_json(msg=msg)

            # First determine if the requested timezone is valid by looking in the zoneinfo
            #  directory.
            zonefile = f"/usr/share/lib/zoneinfo/{value}"
            try:
                if not os.path.isfile(zonefile):
                    self.module.fail_json(msg=f"{value} is not a recognized timezone.")
            except Exception:
                self.module.fail_json(msg=f"Failed to check {zonefile}.")

            # Now set the TZ using chtz
            cmd = ["chtz", value]
            (rc, stdout, stderr) = self.module.run_command(cmd)

            if rc != 0:
                self.module.fail_json(msg=stderr)

            # The best condition check we can do is to check the value of TZ after making the
            #  change.
            TZ = self.__get_timezone()
            if value != TZ:
                msg = f"TZ value does not match post-change (Actual: {TZ}, Expected: {value})."
                self.module.fail_json(msg=msg)

        else:
            self.module.fail_json(msg=f"{key} is not a supported option on target platform")


def main():
    # Construct 'module' and 'tz'
    module = AnsibleModule(
        argument_spec=dict(
            hwclock=dict(type="str", choices=["local", "UTC"], aliases=["rtc"]),
            name=dict(type="str"),
        ),
        required_one_of=[["hwclock", "name"]],
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}
    tz = Timezone(module)

    # Check the current state
    tz.check(phase="before")
    if module.check_mode:
        diff = tz.diff("before", "planned")
        # In check mode, 'planned' state is treated as 'after' state
        diff["after"] = diff.pop("planned")
    else:
        # Make change
        tz.change()
        # Check the current state
        tz.check(phase="after")
        # Examine if the current state matches planned state
        (after, planned) = tz.diff("after", "planned").values()
        if after != planned:
            tz.abort(f"still not desired state, though changes have made - planned: {planned}, after: {after}")
        diff = tz.diff("before", "after")

    changed = diff["before"] != diff["after"]
    if len(tz.msg) > 0:
        module.exit_json(changed=changed, diff=diff, msg="\n".join(tz.msg))
    else:
        module.exit_json(changed=changed, diff=diff)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2025, Tom Hesse <contact@tomhesse.xyz>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: zpool
short_description: Manage ZFS zpools
version_added: 11.0.0
description:
  - Create, destroy, and modify ZFS zpools and their vdev layouts, pool properties, and filesystem properties.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: partial
    details:
      - In check mode, any C(zpool) subcommand that supports the dry-run flag (C(-n)) is run with C(-n) and its simulated
        output is included in the module's diff results.
  diff_mode:
    support: full
author:
  - Tom Hesse (@tomhesse)
options:
  name:
    description:
      - Name of the zpool to manage.
    required: true
    type: str
  state:
    description:
      - Whether the pool should exist.
    choices: [present, absent]
    default: present
    type: str
  disable_new_features:
    description:
      - If V(true), disable new ZFS feature flags when creating.
    type: bool
    default: false
  force:
    description:
      - If V(true), force operations (for example overwrite existing devices).
    type: bool
    default: false
  pool_properties:
    description:
      - Dictionary of ZFS pool properties to set (for example V(autoexpand), V(cachefile)).
    type: dict
    default: {}
  filesystem_properties:
    description:
      - Dictionary of ZFS filesystem properties to set on the root dataset (for example V(compression), V(dedup)).
    type: dict
    default: {}
  mountpoint:
    description:
      - Filesystem mountpoint for the root dataset.
    type: str
  altroot:
    description:
      - Alternate root for mounting filesystems.
    type: str
  temp_name:
    description:
      - Temporary name used during pool creation.
    type: str
  vdevs:
    description:
      - List of vdev definitions for the pool.
    type: list
    elements: dict
    suboptions:
      role:
        description:
          - Special vdev role (for example V(log), V(cache), V(spare)).
        type: str
        choices: [log, cache, spare, dedup, special]
      type:
        description:
          - Vdev topology (for example V(stripe), V(mirror), V(raidz)).
        type: str
        choices: [stripe, mirror, raidz, raidz1, raidz2, raidz3]
        default: stripe
      disks:
        description:
          - List of device paths to include in this vdev.
        required: true
        type: list
        elements: path
"""

EXAMPLES = r"""
- name: Create pool "tank" on /dev/sda
  community.general.zpool:
    name: tank
    vdevs:
      - disks:
          - /dev/sda

- name: Create mirrored pool "tank"
  community.general.zpool:
    name: tank
    vdevs:
      - type: mirror
        disks:
          - /dev/sda
          - /dev/sdb

- name: Add a cache device to tank
  community.general.zpool:
    name: tank
    vdevs:
      - disks:
          - /dev/sda
      - role: cache
        disks:
          - /dev/nvme0n1

- name: Set pool and filesystem properties
  community.general.zpool:
    name: tank
    ashift: 12
    compression: lz4
    vdevs:
      - disks:
          - /dev/sda

- name: Destroy pool "tank"
  community.general.zpool:
    name: tank
    state: absent
"""

import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt


class Zpool:
    def __init__(
        self,
        module,
        name,
        disable_new_features,
        force,
        pool_properties,
        filesystem_properties,
        mountpoint,
        altroot,
        temp_name,
        vdevs,
    ):
        self.module = module
        self.name = name
        self.disable_new_features = disable_new_features
        self.force = force
        self.pool_properties = pool_properties
        self.filesystem_properties = filesystem_properties
        self.mountpoint = mountpoint
        self.altroot = altroot
        self.temp_name = temp_name
        self.vdevs = vdevs
        self.zpool_cmd = module.get_bin_path("zpool", required=True)
        self.zfs_cmd = module.get_bin_path("zfs", required=True)
        self.changed = False

        self.zpool_runner = CmdRunner(
            module,
            command=self.zpool_cmd,
            arg_formats=dict(
                subcommand=cmd_runner_fmt.as_list(),
                disable_new_features=cmd_runner_fmt.as_bool("-d"),
                force=cmd_runner_fmt.as_bool("-f"),
                dry_run=cmd_runner_fmt.as_bool("-n"),
                pool_properties=cmd_runner_fmt.as_func(
                    lambda props: sum([["-o", f"{prop}={value}"] for prop, value in (props or {}).items()], [])
                ),
                filesystem_properties=cmd_runner_fmt.as_func(
                    lambda props: sum([["-O", f"{prop}={value}"] for prop, value in (props or {}).items()], [])
                ),
                mountpoint=cmd_runner_fmt.as_opt_val("-m"),
                altroot=cmd_runner_fmt.as_opt_val("-R"),
                temp_name=cmd_runner_fmt.as_opt_val("-t"),
                name=cmd_runner_fmt.as_list(),
                vdevs=cmd_runner_fmt.as_func(
                    lambda vdevs: sum(
                        [
                            ([vdev["role"]] if vdev.get("role") else [])
                            + ([] if vdev.get("type", "stripe") == "stripe" else [vdev["type"]])
                            + vdev.get("disks", [])
                            for vdev in (vdevs or [])
                        ],
                        [],
                    )
                ),
                vdev_name=cmd_runner_fmt.as_list(),
                scripted=cmd_runner_fmt.as_bool("-H"),
                parsable=cmd_runner_fmt.as_bool("-p"),
                columns=cmd_runner_fmt.as_opt_val("-o"),
                properties=cmd_runner_fmt.as_list(),
                assignment=cmd_runner_fmt.as_list(),
                full_paths=cmd_runner_fmt.as_bool("-P"),
                real_paths=cmd_runner_fmt.as_bool("-L"),
            ),
        )

        self.zfs_runner = CmdRunner(
            module,
            command=self.zfs_cmd,
            arg_formats=dict(
                subcommand=cmd_runner_fmt.as_list(),
                scripted=cmd_runner_fmt.as_bool("-H"),
                columns=cmd_runner_fmt.as_opt_val("-o"),
                properties=cmd_runner_fmt.as_list(),
                assignment=cmd_runner_fmt.as_list(),
                name=cmd_runner_fmt.as_list(),
            ),
        )

    def exists(self):
        with self.zpool_runner("subcommand name") as ctx:
            rc, stdout, stderr = ctx.run(subcommand="list", name=self.name)
        return rc == 0

    def create(self):
        with self.zpool_runner(
            "subcommand disable_new_features force dry_run pool_properties filesystem_properties mountpoint altroot temp_name name vdevs",
            check_rc=True,
        ) as ctx:
            rc, stdout, stderr = ctx.run(subcommand="create", dry_run=self.module.check_mode)
        self.changed = True
        if self.module.check_mode:
            return {"prepared": stdout}

    def destroy(self):
        if self.module.check_mode:
            self.changed = True
            return
        with self.zpool_runner("subcommand name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(subcommand="destroy")
        self.changed = True

    def list_pool_properties(self):
        with self.zpool_runner("subcommand scripted columns properties name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(
                subcommand="get",
                scripted=True,
                columns="property,value",
                properties="all",
            )

        props = {}
        for line in stdout.splitlines():
            prop, value = line.split("\t", 1)
            props[prop] = value
        return props

    def set_pool_properties_if_changed(self):
        current = self.list_pool_properties()
        before = {}
        after = {}
        for prop, value in self.pool_properties.items():
            if current.get(prop) != str(value):
                before[prop] = current.get(prop)
                if not self.module.check_mode:
                    with self.zpool_runner("subcommand assignment name", check_rc=True) as ctx:
                        rc, stdout, stderr = ctx.run(subcommand="set", assignment=f"{prop}={value}")
                after[prop] = str(value)
                self.changed = True
        return {"before": {"pool_properties": before}, "after": {"pool_properties": after}}

    def list_filesystem_properties(self):
        with self.zfs_runner("subcommand scripted columns properties name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(
                subcommand="get",
                scripted=True,
                columns="property,value",
                properties="all",
            )

        props = {}
        for line in stdout.splitlines():
            prop, value = line.split("\t", 1)
            props[prop] = value
        return props

    def set_filesystem_properties_if_changed(self):
        current = self.list_filesystem_properties()
        before = {}
        after = {}
        for prop, value in self.filesystem_properties.items():
            if current.get(prop) != str(value):
                before[prop] = current.get(prop)
                if not self.module.check_mode:
                    with self.zfs_runner("subcommand assignment name", check_rc=True) as ctx:
                        rc, stdout, stderr = ctx.run(subcommand="set", assignment=f"{prop}={value}")
                after[prop] = str(value)
                self.changed = True
        return {"before": {"filesystem_properties": before}, "after": {"filesystem_properties": after}}

    def base_device(self, device):
        if not device.startswith("/dev/"):
            return device

        # loop devices
        match = re.match(r"^(/dev/loop\d+)$", device)
        if match:
            return match.group(1)

        # nvme drives
        match = re.match(r"^(.*?)(p\d+)$", device)
        if match:
            return match.group(1)

        # sata/scsi drives
        match = re.match(r"^(/dev/(?:sd|vd)[a-z])\d+$", device)
        if match:
            return match.group(1)

        return device

    def get_current_layout(self):
        with self.zpool_runner("subcommand full_paths real_paths name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(subcommand="status", full_paths=True, real_paths=True)

        vdevs = []
        current = None
        in_config = False

        def flush_current(current):
            if current:
                if current.get("role") is None:
                    current.pop("role", None)
                vdevs.append(current)
            return None

        for line in stdout.splitlines():
            if not in_config:
                if line.strip().startswith("config:"):
                    in_config = True
                continue

            if not line.strip() or line.strip().startswith("NAME"):
                continue

            partitions = line.split()
            device = partitions[0]

            if device == self.name:
                continue

            if device in ("logs", "cache", "spares"):
                current = flush_current(current)
                role = "spare" if device == "spares" else device.rstrip("s")
                current = {"role": role, "type": None, "disks": []}
                continue

            match_group = re.match(r"^(mirror|raidz\d?)-\d+$", device)
            if match_group:
                if current and current.get("type") is not None:
                    current = flush_current(current)
                kind = match_group.group(1)
                role = current.get("role") if current and current.get("type") is None else None
                current = {"role": role, "type": kind, "disks": []}
                continue

            if device.startswith("/"):
                base_device = self.base_device(device)
                if current:
                    if current.get("type") is None:
                        entry = {"type": "stripe", "disks": [base_device]}
                        if current.get("role"):
                            entry["role"] = current["role"]
                        vdevs.append(entry)
                        current = None
                    else:
                        current["disks"].append(base_device)
                else:
                    vdevs.append({"type": "stripe", "disks": [base_device]})
                continue

        if current and current.get("type") is not None:
            current = flush_current(current)

        return vdevs

    def normalize_vdevs(self, vdevs):
        alias = {"raidz": "raidz1"}
        normalized = []
        for vdev in vdevs:
            normalized_type = alias.get(vdev.get("type", "stripe"), vdev.get("type", "stripe"))
            entry = {
                "type": normalized_type,
                "disks": sorted(vdev["disks"]),
            }
            role = vdev.get("role")
            if role is not None:
                entry["role"] = role
            normalized.append(entry)
        return sorted(normalized, key=lambda x: (x.get("role", ""), x["type"], x["disks"]))

    def diff_layout(self):
        current = self.normalize_vdevs(self.get_current_layout())
        desired = self.normalize_vdevs(self.vdevs)

        before = {"vdevs": current}
        after = {"vdevs": desired}

        if current != desired:
            self.changed = True

        return {"before": before, "after": after}

    def add_vdevs(self):
        invalid_properties = [k for k in self.pool_properties if k != "ashift"]
        if invalid_properties:
            self.module.warn(f"zpool add only supports 'ashift', ignoring: {invalid_properties}")

        diff = self.diff_layout()
        before_vdevs = diff["before"]["vdevs"]
        after_vdevs = diff["after"]["vdevs"]

        to_add = [vdev for vdev in after_vdevs if vdev not in before_vdevs]
        if not to_add:
            return {}

        with self.zpool_runner("subcommand force dry_run pool_properties name vdevs", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(
                subcommand="add",
                dry_run=self.module.check_mode,
                pool_properties={"ashift": self.pool_properties["ashift"]} if "ashift" in self.pool_properties else {},
                vdevs=to_add,
            )

        self.changed = True
        if self.module.check_mode:
            return {"prepared": stdout}

    def list_vdevs_with_names(self):
        with self.zpool_runner("subcommand full_paths real_paths name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(subcommand="status", full_paths=True, real_paths=True)
        in_cfg = False
        saw_pool = False
        vdevs = []
        current = None
        for line in stdout.splitlines():
            if not in_cfg:
                if line.strip().startswith("config:"):
                    in_cfg = True
                continue
            if not line.strip() or line.strip().startswith("NAME"):
                continue
            partitions = line.strip().split()
            device = partitions[0]
            if not saw_pool:
                if device == self.name:
                    saw_pool = True
                continue
            if re.match(r"^(mirror|raidz\d?)\-\d+$", device) or device in ("cache", "logs", "spares"):
                if current:
                    vdevs.append(current)
                vdev_type = (
                    "stripe"
                    if device in ("cache", "logs", "spares")
                    else ("mirror" if device.startswith("mirror") else "raidz")
                )
                current = {"name": device, "type": vdev_type, "disks": []}
                continue
            if device.startswith("/") and current:
                current["disks"].append(self.base_device(device))
                continue
            if device.startswith("/"):
                base_device = self.base_device(device)
                vdevs.append({"name": base_device, "type": "stripe", "disks": [base_device]})
        if current:
            vdevs.append(current)
        return vdevs

    def remove_vdevs(self):
        current = self.list_vdevs_with_names()
        current_disks = {disk for vdev in current for disk in vdev["disks"]}
        desired_disks = {disk for vdev in self.vdevs for disk in vdev.get("disks", [])}
        gone = current_disks - desired_disks
        to_remove = [vdev["name"] for vdev in current if any(disk in gone for disk in vdev["disks"])]
        if not to_remove:
            return {}
        with self.zpool_runner("subcommand dry_run name vdev_name", check_rc=True) as ctx:
            rc, stdout, stderr = ctx.run(subcommand="remove", dry_run=self.module.check_mode, vdev_name=to_remove)
        self.changed = True
        if self.module.check_mode:
            return {"prepared": stdout}
        before = [vdev["name"] for vdev in current]
        after = [name for name in before if name not in to_remove]
        return {"before": {"vdevs": before}, "after": {"vdevs": after}}


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            state=dict(type="str", choices=["present", "absent"], default="present"),
            disable_new_features=dict(type="bool", default=False),
            force=dict(type="bool", default=False),
            pool_properties=dict(type="dict", default={}),
            filesystem_properties=dict(type="dict", default={}),
            mountpoint=dict(type="str"),
            altroot=dict(type="str"),
            temp_name=dict(type="str"),
            vdevs=dict(
                type="list",
                elements="dict",
                options=dict(
                    role=dict(
                        type="str",
                        choices=["log", "cache", "spare", "dedup", "special"],
                    ),
                    type=dict(
                        type="str",
                        choices=["stripe", "mirror", "raidz", "raidz1", "raidz2", "raidz3"],
                        default="stripe",
                    ),
                    disks=dict(
                        type="list",
                        elements="path",
                        required=True,
                    ),
                ),
            ),
        ),
        supports_check_mode=True,
        required_if=[("state", "present", ["vdevs"])],
    )

    name = module.params.get("name")
    state = module.params.get("state")
    disable_new_features = module.params.get("disable_new_features")
    force = module.params.get("force")
    pool_properties = module.params.get("pool_properties")
    filesystem_properties = module.params.get("filesystem_properties")
    mountpoint = module.params.get("mountpoint")
    altroot = module.params.get("altroot")
    temp_name = module.params.get("temp_name")
    vdevs = module.params.get("vdevs")

    for property_key in ("pool_properties", "filesystem_properties"):
        for key, value in list(module.params.get(property_key, {}).items()):
            if isinstance(value, bool):
                module.params[property_key][key] = "on" if value else "off"

    if state != "absent":
        for idx, vdev in enumerate(vdevs, start=1):
            disks = vdev.get("disks")
            if not isinstance(disks, list) or len(disks) == 0:
                module.fail_json(msg=f"vdev #{idx}: at least one disk is required (got: {disks!r})")

    result = dict(
        name=name,
        state=state,
    )

    zpool = Zpool(
        module,
        name,
        disable_new_features,
        force,
        pool_properties,
        filesystem_properties,
        mountpoint,
        altroot,
        temp_name,
        vdevs,
    )

    if state == "present":
        if zpool.exists():
            vdev_layout_diff = zpool.diff_layout()

            add_vdev_diff = zpool.add_vdevs() or {}
            remove_vdev_diff = zpool.remove_vdevs() or {}
            pool_properties_diff = zpool.set_pool_properties_if_changed()
            filesystem_properties_diff = zpool.set_filesystem_properties_if_changed()

            before = {}
            after = {}
            for diff in (vdev_layout_diff, pool_properties_diff, filesystem_properties_diff):
                before.update(diff.get("before", {}))
                after.update(diff.get("after", {}))

            result["diff"] = {"before": before, "after": after}

            if module.check_mode:
                prepared = ""
                for diff in (add_vdev_diff, remove_vdev_diff):
                    if "prepared" in diff:
                        prepared += diff["prepared"] if not prepared else f"\n{diff['prepared']}"
                result["diff"]["prepared"] = prepared
        else:
            if module.check_mode:
                result["diff"] = zpool.create()
            else:
                before_vdevs = []
                desired_vdevs = zpool.normalize_vdevs(zpool.vdevs)
                zpool.create()
                result["diff"] = {
                    "before": {"state": "absent", "vdevs": before_vdevs},
                    "after": {"state": state, "vdevs": desired_vdevs},
                }

    elif state == "absent":
        if zpool.exists():
            before_vdevs = zpool.get_current_layout()
            zpool.destroy()
            result["diff"] = {
                "before": {"state": "present", "vdevs": before_vdevs},
                "after": {"state": state, "vdevs": []},
            }
        else:
            result["diff"] = {}

    result["diff"]["before_header"] = name
    result["diff"]["after_header"] = name

    result["changed"] = zpool.changed
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Michael DeHaan <michael@ansible.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: say
short_description: Makes a computer to speak
description:
  - Makes a computer speak! Amuse your friends, annoy your coworkers!
notes:
  - In 2.5, this module has been renamed from C(osx_say) to M(community.general.say).
  - If you like this module, you may also be interested in the osx_say callback plugin.
  - A list of available voices, with language, can be found by running C(say -v ?) on a OSX host and C(espeak --voices) on
    a Linux host.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  msg:
    type: str
    description:
      - What to say.
    required: true
  voice:
    type: str
    description:
      - What voice to use.
requirements: [say or espeak or espeak-ng]
author:
  - "Ansible Core Team"
  - "Michael DeHaan (@mpdehaan)"
"""

EXAMPLES = r"""
- name: Makes a computer to speak
  community.general.say:
    msg: '{{ inventory_hostname }} is all done'
    voice: Zarvox
  delegate_to: localhost
"""
import platform

from ansible.module_utils.basic import AnsibleModule


def say(module, executable, msg, voice):
    cmd = [executable, msg]
    if voice:
        cmd.extend(("-v", voice))
    module.run_command(cmd, check_rc=True)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            msg=dict(required=True),
            voice=dict(),
        ),
        supports_check_mode=True,
    )

    msg = module.params["msg"]
    voice = module.params["voice"]
    possibles = ("say", "espeak", "espeak-ng")

    if platform.system() != "Darwin":
        # 'say' binary available, it might be GNUstep tool which doesn't support 'voice' parameter
        voice = None

    for possible in possibles:
        executable = module.get_bin_path(possible)
        if executable:
            break
    else:
        module.fail_json(msg=f"Unable to find either {', '.join(possibles)}")

    if module.check_mode:
        module.exit_json(msg=msg, changed=False)

    say(module, executable, msg, voice)

    module.exit_json(msg=msg, changed=True)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Raul Melo
# Written by Raul Melo <raulmelo@gmail.com>
# Based on yum module written by Seth Vidal <skvidal at fedoraproject.org>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: swdepot
short_description: Manage packages with swdepot package manager (HP-UX)
description:
  - Installs, upgrades, and removes packages with C(swdepot) package manager (HP-UX).
notes: []
author: "Raul Melo (@melodous)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Package name.
    aliases: [pkg]
    required: true
    type: str
  state:
    description:
      - Whether to install (V(present), V(latest)), or remove (V(absent)) a package.
    required: true
    choices: ['present', 'latest', 'absent']
    type: str
  depot:
    description:
      - The source repository from which install or upgrade a package.
    type: str
"""

EXAMPLES = r"""
- name: Install a package
  community.general.swdepot:
    name: unzip-6.0
    state: present
    depot: 'repository:/path'

- name: Install the latest version of a package
  community.general.swdepot:
    name: unzip
    state: latest
    depot: 'repository:/path'

- name: Remove a package
  community.general.swdepot:
    name: unzip
    state: absent
"""

import re

from ansible.module_utils.basic import AnsibleModule


def compare_package(version1, version2):
    """Compare version packages.
    Return values:
    -1 first minor
    0 equal
    1 first greater"""

    def normalize(v):
        return [int(x) for x in re.sub(r"(\.0+)*$", "", v).split(".")]

    normalized_version1 = normalize(version1)
    normalized_version2 = normalize(version2)
    if normalized_version1 == normalized_version2:
        rc = 0
    elif normalized_version1 < normalized_version2:
        rc = -1
    else:
        rc = 1
    return rc


def query_package(module, name, depot=None):
    """Returns whether a package is installed or not and version."""

    cmd_list = ["/usr/sbin/swlist", "-a", "revision", "-l", "product"]
    if depot:
        cmd_list.extend(["-s", depot])
    cmd_list.append(name)
    rc, stdout, stderr = module.run_command(cmd_list)
    if rc == 0:
        stdout = "".join(line for line in stdout.splitlines(True) if name in line)
        version = re.sub(r"\s\s+|\t", " ", stdout).strip().split()[1]
    else:
        version = None

    return rc, version


def remove_package(module, name):
    """Uninstall package if installed."""

    cmd_remove = "/usr/sbin/swremove"
    rc, stdout, stderr = module.run_command([cmd_remove, name])

    if rc == 0:
        return rc, stdout
    else:
        return rc, stderr


def install_package(module, depot, name):
    """Install package if not already installed"""

    cmd_install = ["/usr/sbin/swinstall", "-x", "mount_all_filesystems=false"]
    rc, stdout, stderr = module.run_command(cmd_install + ["-s", depot, name])
    if rc == 0:
        return rc, stdout
    else:
        return rc, stderr


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(aliases=["pkg"], required=True),
            state=dict(choices=["present", "absent", "latest"], required=True),
            depot=dict(),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}
    name = module.params["name"]
    state = module.params["state"]
    depot = module.params["depot"]

    changed = False
    msg = "No changed"
    rc = 0
    if (state == "present" or state == "latest") and depot is None:
        output = "depot parameter is mandatory in present or latest task"
        module.fail_json(name=name, msg=output, rc=rc)

    # Check local version
    rc, version_installed = query_package(module, name)
    if not rc:
        installed = True
        msg = "Already installed"

    else:
        installed = False

    if (state == "present" or state == "latest") and installed is False:
        if module.check_mode:
            module.exit_json(changed=True)
        rc, output = install_package(module, depot, name)

        if not rc:
            changed = True
            msg = "Package installed"

        else:
            module.fail_json(name=name, msg=output, rc=rc)

    elif state == "latest" and installed is True:
        # Check depot version
        rc, version_depot = query_package(module, name, depot)

        if not rc:
            if compare_package(version_installed, version_depot) == -1:
                if module.check_mode:
                    module.exit_json(changed=True)
                # Install new version
                rc, output = install_package(module, depot, name)

                if not rc:
                    msg = f"Package upgraded, Before {version_installed} Now {version_depot}"
                    changed = True

                else:
                    module.fail_json(name=name, msg=output, rc=rc)

        else:
            output = f"Software package not in repository {depot}"
            module.fail_json(name=name, msg=output, rc=rc)

    elif state == "absent" and installed is True:
        if module.check_mode:
            module.exit_json(changed=True)
        rc, output = remove_package(module, name)
        if not rc:
            changed = True
            msg = "Package removed"
        else:
            module.fail_json(name=name, msg=output, rc=rc)

    if module.check_mode:
        module.exit_json(changed=False)

    module.exit_json(changed=changed, name=name, state=state, msg=msg)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016-2017 Hewlett Packard Enterprise Development LP
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: oneview_ethernet_network
short_description: Manage OneView Ethernet Network resources
description:
  - Provides an interface to manage Ethernet Network resources. Can create, update, or delete.
requirements:
  - hpOneView >= 3.1.0
author:
  - Felipe Bulsoni (@fgbulsoni)
  - Thiago Miotto (@tmiotto)
  - Adriane Cardozo (@adriane-cardozo)
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - Indicates the desired state for the Ethernet Network resource.
      - V(present) ensures data properties are compliant with OneView.
      - V(absent) removes the resource from OneView, if it exists.
      - V(default_bandwidth_reset) resets the network connection template to the default.
    type: str
    default: present
    choices: [present, absent, default_bandwidth_reset]
  data:
    description:
      - List with Ethernet Network properties.
    type: dict
    required: true
extends_documentation_fragment:
  - community.general._oneview
  - community.general._oneview.validateetag
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Ensure that the Ethernet Network is present using the default configuration
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: present
    data:
      name: 'Test Ethernet Network'
      vlanId: '201'
  delegate_to: localhost

- name: Update the Ethernet Network changing bandwidth and purpose
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: present
    data:
      name: 'Test Ethernet Network'
      purpose: Management
      bandwidth:
        maximumBandwidth: 3000
        typicalBandwidth: 2000
  delegate_to: localhost

- name: Ensure that the Ethernet Network is present with name 'Renamed Ethernet Network'
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: present
    data:
      name: 'Test Ethernet Network'
      newName: 'Renamed Ethernet Network'
  delegate_to: localhost

- name: Ensure that the Ethernet Network is absent
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: absent
    data:
      name: 'New Ethernet Network'
  delegate_to: localhost

- name: Create Ethernet networks in bulk
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: present
    data:
      vlanIdRange: '1-10,15,17'
      purpose: General
      namePrefix: TestNetwork
      smartLink: false
      privateNetwork: false
      bandwidth:
        maximumBandwidth: 10000
        typicalBandwidth: 2000
  delegate_to: localhost

- name: Reset to the default network connection template
  community.general.oneview_ethernet_network:
    config: '/etc/oneview/oneview_config.json'
    state: default_bandwidth_reset
    data:
      name: 'Test Ethernet Network'
  delegate_to: localhost
"""

RETURN = r"""
ethernet_network:
  description: Has the facts about the Ethernet Networks.
  returned: On O(state=present). Can be null.
  type: dict

ethernet_network_bulk:
  description: Has the facts about the Ethernet Networks affected by the bulk insert.
  returned: When V(vlanIdRange) attribute is in O(data) argument. Can be null.
  type: dict

ethernet_network_connection_template:
  description: Has the facts about the Ethernet Network Connection Template.
  returned: On O(state=default_bandwidth_reset). Can be null.
  type: dict
"""

from ansible_collections.community.general.plugins.module_utils._oneview import (
    OneViewModuleBase,
    OneViewModuleResourceNotFound,
)


class EthernetNetworkModule(OneViewModuleBase):
    MSG_CREATED = "Ethernet Network created successfully."
    MSG_UPDATED = "Ethernet Network updated successfully."
    MSG_DELETED = "Ethernet Network deleted successfully."
    MSG_ALREADY_PRESENT = "Ethernet Network is already present."
    MSG_ALREADY_ABSENT = "Ethernet Network is already absent."

    MSG_BULK_CREATED = "Ethernet Networks created successfully."
    MSG_MISSING_BULK_CREATED = "Some missing Ethernet Networks were created successfully."
    MSG_BULK_ALREADY_EXIST = "The specified Ethernet Networks already exist."
    MSG_CONNECTION_TEMPLATE_RESET = "Ethernet Network connection template was reset to the default."
    MSG_ETHERNET_NETWORK_NOT_FOUND = "Ethernet Network was not found."

    RESOURCE_FACT_NAME = "ethernet_network"

    def __init__(self):
        argument_spec = dict(
            state=dict(type="str", default="present", choices=["absent", "default_bandwidth_reset", "present"]),
            data=dict(type="dict", required=True),
        )

        super().__init__(additional_arg_spec=argument_spec, validate_etag_support=True)

        self.resource_client = self.oneview_client.ethernet_networks

    def execute_module(self):
        changed, msg, ansible_facts, resource = False, "", {}, None

        if self.data.get("name"):
            resource = self.get_by_name(self.data["name"])

        if self.state == "present":
            if self.data.get("vlanIdRange"):
                return self._bulk_present()
            else:
                return self._present(resource)
        elif self.state == "absent":
            return self.resource_absent(resource)
        elif self.state == "default_bandwidth_reset":
            changed, msg, ansible_facts = self._default_bandwidth_reset(resource)
            return dict(changed=changed, msg=msg, ansible_facts=ansible_facts)

    def _present(self, resource):
        bandwidth = self.data.pop("bandwidth", None)
        scope_uris = self.data.pop("scopeUris", None)
        result = self.resource_present(resource, self.RESOURCE_FACT_NAME)

        if bandwidth:
            if self._update_connection_template(result["ansible_facts"]["ethernet_network"], bandwidth)[0]:
                result["changed"] = True
                result["msg"] = self.MSG_UPDATED

        if scope_uris is not None:
            result = self.resource_scopes_set(result, "ethernet_network", scope_uris)

        return result

    def _bulk_present(self):
        vlan_id_range = self.data["vlanIdRange"]
        result = dict(ansible_facts={})
        ethernet_networks = self.resource_client.get_range(self.data["namePrefix"], vlan_id_range)

        if not ethernet_networks:
            self.resource_client.create_bulk(self.data)
            result["changed"] = True
            result["msg"] = self.MSG_BULK_CREATED

        else:
            vlan_ids = self.resource_client.dissociate_values_or_ranges(vlan_id_range)
            for net in ethernet_networks[:]:
                vlan_ids.remove(net["vlanId"])

            if len(vlan_ids) == 0:
                result["msg"] = self.MSG_BULK_ALREADY_EXIST
                result["changed"] = False
            else:
                if len(vlan_ids) == 1:
                    self.data["vlanIdRange"] = f"{vlan_ids[0]}-{vlan_ids[0]}"
                else:
                    self.data["vlanIdRange"] = ",".join(str(s) for s in vlan_ids)

                self.resource_client.create_bulk(self.data)
                result["changed"] = True
                result["msg"] = self.MSG_MISSING_BULK_CREATED
        result["ansible_facts"]["ethernet_network_bulk"] = self.resource_client.get_range(
            self.data["namePrefix"], vlan_id_range
        )

        return result

    def _update_connection_template(self, ethernet_network, bandwidth):
        if "connectionTemplateUri" not in ethernet_network:
            return False, None

        connection_template = self.oneview_client.connection_templates.get(ethernet_network["connectionTemplateUri"])

        merged_data = connection_template.copy()
        merged_data.update({"bandwidth": bandwidth})

        if not self.compare(connection_template, merged_data):
            connection_template = self.oneview_client.connection_templates.update(merged_data)
            return True, connection_template
        else:
            return False, None

    def _default_bandwidth_reset(self, resource):
        if not resource:
            raise OneViewModuleResourceNotFound(self.MSG_ETHERNET_NETWORK_NOT_FOUND)

        default_connection_template = self.oneview_client.connection_templates.get_default()

        changed, connection_template = self._update_connection_template(
            resource, default_connection_template["bandwidth"]
        )

        return (
            changed,
            self.MSG_CONNECTION_TEMPLATE_RESET,
            dict(ethernet_network_connection_template=connection_template),
        )


def main():
    EthernetNetworkModule().run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Scaleway database backups management module
#
# Copyright (C) 2020 Guillaume Rodriguez (g.rodriguez@opendecide.com).
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: scaleway_database_backup
short_description: Scaleway database backups management module
version_added: 1.2.0
author: Guillaume Rodriguez (@guillaume_ro_fr)
description:
  - This module manages database backups on Scaleway account U(https://developer.scaleway.com).
extends_documentation_fragment:
  - community.general._scaleway
  - community.general._attributes
  - community.general._scaleway.actiongroup_scaleway
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  action_group:
    version_added: 11.3.0

options:
  state:
    description:
      - Indicate desired state of the database backup.
      - V(present) creates a backup.
      - V(absent) deletes the backup.
      - V(exported) creates a download link for the backup.
      - V(restored) restores the backup to a new database.
    type: str
    default: present
    choices:
      - present
      - absent
      - exported
      - restored

  region:
    description:
      - Scaleway region to use (for example V(fr-par)).
    type: str
    required: true
    choices:
      - fr-par
      - nl-ams
      - pl-waw

  id:
    description:
      - UUID used to identify the database backup.
      - Required for V(absent), V(exported) and V(restored) states.
    type: str

  name:
    description:
      - Name used to identify the database backup.
      - Required for V(present) state.
      - Ignored when O(state=absent), O(state=exported) or O(state=restored).
    type: str

  database_name:
    description:
      - Name used to identify the database.
      - Required for V(present) and V(restored) states.
      - Ignored when O(state=absent) or O(state=exported).
    type: str

  instance_id:
    description:
      - UUID of the instance associated to the database backup.
      - Required for V(present) and V(restored) states.
      - Ignored when O(state=absent) or O(state=exported).
    type: str

  expires_at:
    description:
      - Expiration datetime of the database backup (ISO 8601 format).
      - Ignored when O(state=absent), O(state=exported) or O(state=restored).
    type: str

  wait:
    description:
      - Wait for the instance to reach its desired state before returning.
    type: bool
    default: false

  wait_timeout:
    description:
      - Time to wait for the backup to reach the expected state.
    type: int
    default: 300

  wait_sleep_time:
    description:
      - Time to wait before every attempt to check the state of the backup.
    type: int
    default: 3
"""

EXAMPLES = r"""
- name: Create a backup
  community.general.scaleway_database_backup:
    name: 'my_backup'
    state: present
    region: 'fr-par'
    database_name: 'my-database'
    instance_id: '50968a80-2909-4e5c-b1af-a2e19860dddb'

- name: Export a backup
  community.general.scaleway_database_backup:
    id: '6ef1125a-037e-494f-a911-6d9c49a51691'
    state: exported
    region: 'fr-par'

- name: Restore a backup
  community.general.scaleway_database_backup:
    id: '6ef1125a-037e-494f-a911-6d9c49a51691'
    state: restored
    region: 'fr-par'
    database_name: 'my-new-database'
    instance_id: '50968a80-2909-4e5c-b1af-a2e19860dddb'

- name: Remove a backup
  community.general.scaleway_database_backup:
    id: '6ef1125a-037e-494f-a911-6d9c49a51691'
    state: absent
    region: 'fr-par'
"""

RETURN = r"""
metadata:
  description: Backup metadata.
  returned: when O(state=present), O(state=exported), or O(state=restored)
  type: dict
  sample:
    {
      "metadata": {
        "created_at": "2020-08-06T12:42:05.631049Z",
        "database_name": "my-database",
        "download_url": null,
        "download_url_expires_at": null,
        "expires_at": null,
        "id": "a15297bd-0c4a-4b4f-8fbb-b36a35b7eb07",
        "instance_id": "617be32e-6497-4ed7-b4c7-0ee5a81edf49",
        "instance_name": "my-instance",
        "name": "backup_name",
        "region": "fr-par",
        "size": 600000,
        "status": "ready",
        "updated_at": "2020-08-06T12:42:10.581649Z"
      }
    }
"""

import datetime
import time

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._datetime import (
    now,
)
from ansible_collections.community.general.plugins.module_utils._scaleway import (
    SCALEWAY_REGIONS,
    Scaleway,
    scaleway_argument_spec,
)

stable_states = (
    "ready",
    "deleting",
)


def wait_to_complete_state_transition(module, account_api, backup=None):
    wait_timeout = module.params["wait_timeout"]
    wait_sleep_time = module.params["wait_sleep_time"]

    if backup is None or backup["status"] in stable_states:
        return backup

    start = now()
    end = start + datetime.timedelta(seconds=wait_timeout)
    while now() < end:
        module.debug("We are going to wait for the backup to finish its transition")

        response = account_api.get(f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup['id']}")
        if not response.ok:
            module.fail_json(msg=f"Error getting backup [{response.status_code}: {response.json}]")
            break
        response_json = response.json

        if response_json["status"] in stable_states:
            module.debug("It seems that the backup is not in transition anymore.")
            module.debug(f"Backup in state: {response_json['status']}")
            return response_json
        time.sleep(wait_sleep_time)
    else:
        module.fail_json(msg="Backup takes too long to finish its transition")


def present_strategy(module, account_api, backup):
    name = module.params["name"]
    database_name = module.params["database_name"]
    instance_id = module.params["instance_id"]
    expiration_date = module.params["expires_at"]

    if backup is not None:
        if (backup["name"] == name or name is None) and (
            backup["expires_at"] == expiration_date or expiration_date is None
        ):
            wait_to_complete_state_transition(module, account_api, backup)
            module.exit_json(changed=False)

        if module.check_mode:
            module.exit_json(changed=True)

        payload = {}
        if name is not None:
            payload["name"] = name
        if expiration_date is not None:
            payload["expires_at"] = expiration_date

        response = account_api.patch(f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup['id']}", payload)
        if response.ok:
            result = wait_to_complete_state_transition(module, account_api, response.json)
            module.exit_json(changed=True, metadata=result)

        module.fail_json(msg=f"Error modifying backup [{response.status_code}: {response.json}]")

    if module.check_mode:
        module.exit_json(changed=True)

    payload = {"name": name, "database_name": database_name, "instance_id": instance_id}
    if expiration_date is not None:
        payload["expires_at"] = expiration_date

    response = account_api.post(f"/rdb/v1/regions/{module.params.get('region')}/backups", payload)

    if response.ok:
        result = wait_to_complete_state_transition(module, account_api, response.json)
        module.exit_json(changed=True, metadata=result)

    module.fail_json(msg=f"Error creating backup [{response.status_code}: {response.json}]")


def absent_strategy(module, account_api, backup):
    if backup is None:
        module.exit_json(changed=False)

    if module.check_mode:
        module.exit_json(changed=True)

    response = account_api.delete(f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup['id']}")
    if response.ok:
        result = wait_to_complete_state_transition(module, account_api, response.json)
        module.exit_json(changed=True, metadata=result)

    module.fail_json(msg=f"Error deleting backup [{response.status_code}: {response.json}]")


def exported_strategy(module, account_api, backup):
    if backup is None:
        module.fail_json(msg=f'Backup "{module.params["id"]}" not found')

    if backup["download_url"] is not None:
        module.exit_json(changed=False, metadata=backup)

    if module.check_mode:
        module.exit_json(changed=True)

    backup = wait_to_complete_state_transition(module, account_api, backup)
    response = account_api.post(f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup['id']}/export", {})

    if response.ok:
        result = wait_to_complete_state_transition(module, account_api, response.json)
        module.exit_json(changed=True, metadata=result)

    module.fail_json(msg=f"Error exporting backup [{response.status_code}: {response.json}]")


def restored_strategy(module, account_api, backup):
    if backup is None:
        module.fail_json(msg=f'Backup "{module.params["id"]}" not found')

    database_name = module.params["database_name"]
    instance_id = module.params["instance_id"]

    if module.check_mode:
        module.exit_json(changed=True)

    backup = wait_to_complete_state_transition(module, account_api, backup)

    payload = {"database_name": database_name, "instance_id": instance_id}
    response = account_api.post(
        f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup['id']}/restore", payload
    )

    if response.ok:
        result = wait_to_complete_state_transition(module, account_api, response.json)
        module.exit_json(changed=True, metadata=result)

    module.fail_json(msg=f"Error restoring backup [{response.status_code}: {response.json}]")


state_strategy = {
    "present": present_strategy,
    "absent": absent_strategy,
    "exported": exported_strategy,
    "restored": restored_strategy,
}


def core(module):
    state = module.params["state"]
    backup_id = module.params["id"]

    account_api = Scaleway(module)

    if backup_id is None:
        backup_by_id = None
    else:
        response = account_api.get(f"/rdb/v1/regions/{module.params.get('region')}/backups/{backup_id}")
        status_code = response.status_code
        backup_json = response.json
        backup_by_id = None
        if status_code == 404:
            backup_by_id = None
        elif response.ok:
            backup_by_id = backup_json
        else:
            module.fail_json(msg=f"Error getting backup [{status_code}: {response.json['message']}]")

    state_strategy[state](module, account_api, backup_by_id)


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["absent", "present", "exported", "restored"]),
            region=dict(required=True, choices=SCALEWAY_REGIONS),
            id=dict(),
            name=dict(type="str"),
            database_name=dict(),
            instance_id=dict(),
            expires_at=dict(),
            wait=dict(type="bool", default=False),
            wait_timeout=dict(type="int", default=300),
            wait_sleep_time=dict(type="int", default=3),
        )
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_together=[
            ["database_name", "instance_id"],
        ],
        required_if=[
            ["state", "present", ["name", "database_name", "instance_id"]],
            ["state", "absent", ["id"]],
            ["state", "exported", ["id"]],
            ["state", "restored", ["id", "database_name", "instance_id"]],
        ],
    )

    core(module)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Adfinis SyGroup AG
# @keachi
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: udm_group
author:
  - keachi (@keachi)
short_description: Manage of the posix group
description:
  - This module allows to manage user groups on a univention corporate server (UCS). It uses the Python API of the UCS to
    create a new object or edit it.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: partial
options:
  state:
    default: "present"
    choices: [present, absent]
    description:
      - Whether the group is present or not.
    type: str
  name:
    required: true
    description:
      - Name of the POSIX group.
    type: str
  description:
    description:
      - Group description.
    type: str
  position:
    description:
      - Define the whole LDAP position of the group, for example V(cn=g123m-1A,cn=classes,cn=schueler,cn=groups,ou=schule,dc=example,dc=com).
    type: str
    default: ''
  ou:
    description:
      - LDAP OU, for example V(school) for LDAP OU V(ou=school,dc=example,dc=com).
    type: str
    default: ''
  subpath:
    description:
      - Subpath inside the OU, for example V(cn=classes,cn=students,cn=groups).
    type: str
    default: "cn=groups"
"""


EXAMPLES = r"""
- name: Create a POSIX group
  community.general.udm_group:
    name: g123m-1A

# Create a POSIX group with the exact DN
# C(cn=g123m-1A,cn=classes,cn=students,cn=groups,ou=school,dc=school,dc=example,dc=com)
- name: Create a POSIX group with a DN
  community.general.udm_group:
    name: g123m-1A
    subpath: 'cn=classes,cn=students,cn=groups'
    ou: school

# or
- name: Create a POSIX group with a DN
  community.general.udm_group:
    name: g123m-1A
    position: 'cn=classes,cn=students,cn=groups,ou=school,dc=school,dc=example,dc=com'
"""


RETURN = """#"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._univention_umc import (
    base_dn,
    ldap_search,
    umc_module_for_add,
    umc_module_for_edit,
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(required=True, type="str"),
            description=dict(type="str"),
            position=dict(default="", type="str"),
            ou=dict(default="", type="str"),
            subpath=dict(default="cn=groups", type="str"),
            state=dict(default="present", choices=["present", "absent"], type="str"),
        ),
        supports_check_mode=True,
    )
    name = module.params["name"]
    description = module.params["description"]
    position = module.params["position"]
    ou = module.params["ou"]
    subpath = module.params["subpath"]
    state = module.params["state"]
    changed = False
    diff = None

    groups = list(ldap_search(f"(&(objectClass=posixGroup)(cn={name}))", attr=["cn"]))
    if position != "":
        container = position
    else:
        if ou != "":
            ou = f"ou={ou},"
        if subpath != "":
            subpath = f"{subpath},"
        container = f"{subpath}{ou}{base_dn()}"
    group_dn = f"cn={name},{container}"

    exists = bool(len(groups))

    if state == "present":
        try:
            if not exists:
                grp = umc_module_for_add("groups/group", container)
            else:
                grp = umc_module_for_edit("groups/group", group_dn)
            grp["name"] = name
            grp["description"] = description
            diff = grp.diff()
            changed = grp.diff() != []
            if not module.check_mode:
                if not exists:
                    grp.create()
                else:
                    grp.modify()
        except Exception:
            module.fail_json(msg=f"Creating/editing group {name} in {container} failed")

    if state == "absent" and exists:
        try:
            grp = umc_module_for_edit("groups/group", group_dn)
            if not module.check_mode:
                grp.remove()
            changed = True
        except Exception:
            module.fail_json(msg=f"Removing group {name} failed")

    module.exit_json(changed=changed, name=name, diff=diff, container=container)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2018, Stephan Schwarz <stearz@gmx.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: utm_ca_host_key_cert

author:
  - Stephan Schwarz (@stearz)

short_description: Create, update or destroy ca host_key_cert entry in Sophos UTM

description:
  - Create, update or destroy a ca host_key_cert entry in SOPHOS UTM.
  - This module needs to have the REST Ability of the UTM to be activated.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  name:
    description:
      - The name of the object that identifies the entry.
    required: true
    type: str
  ca:
    description:
      - A reference to an existing utm_ca_signing_ca or utm_ca_verification_ca object.
    required: true
    type: str
  meta:
    description:
      - A reference to an existing utm_ca_meta_x509 object.
    required: true
    type: str
  certificate:
    description:
      - The certificate in PEM format.
    required: true
    type: str
  comment:
    description:
      - Optional comment string.
    type: str
  encrypted:
    description:
      - Optionally enable encryption.
    default: false
    type: bool
  key:
    description:
      - Optional private key in PEM format.
    type: str

extends_documentation_fragment:
  - community.general._utm
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Create a ca_host_key_cert entry
  community.general.utm_ca_host_key_cert:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestHostKeyCertEntry
    ca: REF_ca/signing_ca_OBJECT_STRING
    meta: REF_ca/meta_x509_OBJECT_STRING
    certificate: |
      --- BEGIN CERTIFICATE ---
      . . .
       . . .
      . . .
      --- END CERTIFICATE ---
    state: present

- name: Remove a ca_host_key_cert entry
  community.general.utm_ca_host_key_cert:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestHostKeyCertEntry
    state: absent

- name: Read a ca_host_key_cert entry
  community.general.utm_ca_host_key_cert:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestHostKeyCertEntry
    state: info
"""

RETURN = r"""
result:
  description: The utm object that was created.
  returned: success
  type: complex
  contains:
    _ref:
      description: The reference name of the object.
      type: str
    _locked:
      description: Whether or not the object is currently locked.
      type: bool
    _type:
      description: The type of the object.
      type: str
    name:
      description: The name of the object.
      type: str
    ca:
      description: A reference to an existing utm_ca_signing_ca or utm_ca_verification_ca object.
      type: str
    meta:
      description: A reference to an existing utm_ca_meta_x509 object.
      type: str
    certificate:
      description: The certificate in PEM format.
      type: str
    comment:
      description: Comment string (may be empty string).
      type: str
    encrypted:
      description: If encryption is enabled.
      type: bool
    key:
      description: Private key in PEM format (may be empty string).
      type: str
"""

from ansible_collections.community.general.plugins.module_utils._utm_utils import UTM, UTMModule


def main():
    endpoint = "ca/host_key_cert"
    key_to_check_for_changes = ["ca", "certificate", "comment", "encrypted", "key", "meta"]
    module = UTMModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            ca=dict(type="str", required=True),
            meta=dict(type="str", required=True),
            certificate=dict(type="str", required=True),
            comment=dict(type="str"),
            encrypted=dict(type="bool", default=False),
            key=dict(type="str", no_log=True),
        )
    )
    try:
        # This is needed because the bool value only accepts int values in the backend
        UTM(module, endpoint, key_to_check_for_changes).execute()
    except Exception as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2015, Adam Števko <adam.stevko@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: smartos_image_info
short_description: Get SmartOS image details
description:
  - Retrieve information about all installed images on SmartOS.
author: Adam Števko (@xen0l)
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module
attributes:
  check_mode:
    version_added: 3.3.0
    # This was backported to 2.5.4 and 1.3.11 as well, since this was a bugfix
options:
  filters:
    description:
      - Criteria for selecting image. Can be any value from image manifest and V(published_date), V(published), V(source),
        V(clones), and V(size).
      - More information can be found at U(https://smartos.org/man/8/imgadm) under C(imgadm list).
    type: str
seealso:
  - name: imgadm(8)
    description: Complete manual page for the command C(imgadm).
    link: https://smartos.org/man/8/imgadm
"""

EXAMPLES = r"""
- name: Return information about all installed images
  community.general.smartos_image_info:
  register: result

- name: Return all private active Linux images
  community.general.smartos_image_info:
    filters: "os=linux state=active public=false"
  register: result

- name: Show, how many clones does every image have
  community.general.smartos_image_info:
  register: result

- name: Print information
  ansible.builtin.debug:
    msg: >-
      {{
        result.smartos_images[item]['name'] }}-{{ result.smartos_images[item]['version'] }}
        has {{ result.smartos_images[item]['clones']
      }} VM(s)
  with_items: "{{ result.smartos_images.keys() | list }}"

- name: Print information
  ansible.builtin.debug:
    msg: >-
      {{
        smartos_images[item]['name'] }}-{{ smartos_images[item]['version'] }}
        has {{ smartos_images[item]['clones']
      }} VM(s)
  with_items: "{{ smartos_images.keys() | list }}"
"""

RETURN = r"""
"""

import json

from ansible.module_utils.basic import AnsibleModule


class ImageFacts:
    def __init__(self, module):
        self.module = module

        self.filters = module.params["filters"]

    def return_all_installed_images(self):
        cmd = [self.module.get_bin_path("imgadm"), "list", "-j"]

        if self.filters:
            cmd.append(self.filters)

        (rc, out, err) = self.module.run_command(cmd)

        if rc != 0:
            self.module.exit_json(msg="Failed to get all installed images", stderr=err)

        images = json.loads(out)

        result = {}
        for image in images:
            result[image["manifest"]["uuid"]] = image["manifest"]
            # Merge additional attributes with the image manifest.
            for attrib in ["clones", "source", "zpool"]:
                result[image["manifest"]["uuid"]][attrib] = image[attrib]

        return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            filters=dict(),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    image_facts = ImageFacts(module)

    data = dict(smartos_images=image_facts.return_all_installed_images())

    module.exit_json(**data)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, berenddeboer
# Written by berenddeboer <berend@pobox.com>
# Based on pkgng module written by bleader <bleader at ratonland.org>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: portinstall
short_description: Installing packages from FreeBSD's ports system
description:
  - Manage packages for FreeBSD using C(portinstall).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of package to install/remove.
    aliases: [pkg]
    required: true
    type: str
  state:
    description:
      - State of the package.
    choices: ['present', 'absent']
    default: present
    type: str
  use_packages:
    description:
      - Use packages instead of ports whenever available.
    type: bool
    default: true
author: "berenddeboer (@berenddeboer)"
"""

EXAMPLES = r"""
- name: Install package foo
  community.general.portinstall:
    name: foo
    state: present

- name: Install package security/cyrus-sasl2-saslauthd
  community.general.portinstall:
    name: security/cyrus-sasl2-saslauthd
    state: present

- name: Remove packages foo and bar
  community.general.portinstall:
    name: foo,bar
    state: absent
"""

import re
from shlex import quote as shlex_quote

from ansible.module_utils.basic import AnsibleModule


def query_package(module, name):
    pkg_info_path = module.get_bin_path("pkg_info", False)

    # Assume that if we have pkg_info, we haven't upgraded to pkgng
    if pkg_info_path:
        module.get_bin_path("pkg_glob", True)
        # TODO: convert run_comand() argument to list!
        rc, out, err = module.run_command(f"{pkg_info_path} -e `pkg_glob {shlex_quote(name)}`", use_unsafe_shell=True)
        pkg_info_path = [pkg_info_path]
    else:
        pkg_info_path = [module.get_bin_path("pkg", True), "info"]
        rc, out, err = module.run_command(pkg_info_path + [name])

    found = rc == 0

    if not found:
        # databases/mysql55-client installs as mysql-client, so try solving
        # that the ugly way. Pity FreeBSD doesn't have a fool proof way of checking
        # some package is installed
        name_without_digits = re.sub("[0-9]", "", name)
        if name != name_without_digits:
            rc, out, err = module.run_command(pkg_info_path + [name_without_digits])

        found = rc == 0

    return found


def matching_packages(module, name):
    ports_glob_path = module.get_bin_path("ports_glob", True)
    rc, out, err = module.run_command([ports_glob_path, name])
    # counts the number of packages found
    occurrences = out.count("\n")
    if occurrences == 0:
        name_without_digits = re.sub("[0-9]", "", name)
        if name != name_without_digits:
            rc, out, err = module.run_command([ports_glob_path, name_without_digits])
            occurrences = out.count("\n")
    return occurrences


def remove_packages(module, packages):
    remove_c = 0
    pkg_glob_path = module.get_bin_path("pkg_glob", True)

    # If pkg_delete not found, we assume pkgng
    pkg_delete_path = module.get_bin_path("pkg_delete", False)
    if not pkg_delete_path:
        pkg_delete_path = module.get_bin_path("pkg", True)
        pkg_delete_path = f"{pkg_delete_path} delete -y"

    # Using a for loop in case of error, we can report the package that failed
    for package in packages:
        # Query the package first, to see if we even need to remove
        if not query_package(module, package):
            continue

        # TODO: convert run_comand() argument to list!
        rc, out, err = module.run_command(
            f"{pkg_delete_path} `{pkg_glob_path} {shlex_quote(package)}`", use_unsafe_shell=True
        )

        if query_package(module, package):
            name_without_digits = re.sub("[0-9]", "", package)
            # TODO: convert run_comand() argument to list!
            rc, out, err = module.run_command(
                f"{pkg_delete_path} `{pkg_glob_path} {shlex_quote(name_without_digits)}`", use_unsafe_shell=True
            )
            if query_package(module, package):
                module.fail_json(msg=f"failed to remove {package}: {out}")

        remove_c += 1

    if remove_c > 0:
        module.exit_json(changed=True, msg=f"removed {remove_c} package(s)")

    module.exit_json(changed=False, msg="package(s) already absent")


def install_packages(module, packages, use_packages):
    install_c = 0

    # If portinstall not found, automagically install
    portinstall_path = module.get_bin_path("portinstall", False)
    if not portinstall_path:
        pkg_path = module.get_bin_path("pkg", False)
        if pkg_path:
            module.run_command([pkg_path, "install", "-y", "portupgrade"])
        portinstall_path = module.get_bin_path("portinstall", True)

    if use_packages:
        portinstall_params = ["--use-packages"]
    else:
        portinstall_params = []

    for package in packages:
        if query_package(module, package):
            continue

        # TODO: check how many match
        matches = matching_packages(module, package)
        if matches == 1:
            rc, out, err = module.run_command([portinstall_path, "--batch"] + portinstall_params + [package])
            if not query_package(module, package):
                module.fail_json(msg=f"failed to install {package}: {out}")
        elif matches == 0:
            module.fail_json(msg=f"no matches for package {package}")
        else:
            module.fail_json(msg=f"{matches} matches found for package name {package}")

        install_c += 1

    if install_c > 0:
        module.exit_json(changed=True, msg=f"present {install_c} package(s)")

    module.exit_json(changed=False, msg="package(s) already present")


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"]),
            name=dict(aliases=["pkg"], required=True),
            use_packages=dict(type="bool", default=True),
        )
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    p = module.params

    pkgs = p["name"].split(",")

    if p["state"] == "present":
        install_packages(module, pkgs, p["use_packages"])

    elif p["state"] == "absent":
        remove_packages(module, pkgs)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2021, Andreas Botzner <andreas at botzner dot com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: redis_data_incr
short_description: Increment keys in Redis
version_added: 4.0.0
description:
  - Increment integers or float keys in Redis database and get new value.
  - Default increment for all keys is V(1). For specific increments use the O(increment_int) and O(increment_float) options.
author: "Andreas Botzner (@paginabianca)"
attributes:
  check_mode:
    support: partial
    details:
      - For C(check_mode) to work, the specified O(login_user) needs permission to run the C(GET) command on the key, otherwise
        the module fails.
      - When using C(check_mode) the module tries to calculate the value that Redis would return. If the key is not present,
        V(0.0) is used as value.
  diff_mode:
    support: none
options:
  key:
    description:
      - Database key.
    type: str
    required: true
  increment_int:
    description:
      - Integer amount to increment the key by.
    type: int
  increment_float:
    description:
      - Float amount to increment the key by.
      - This only works with keys that contain float values in their string representation.
    type: float


extends_documentation_fragment:
  - community.general._redis.documentation
  - community.general._attributes

seealso:
  - module: community.general.redis_data
  - module: community.general.redis_data_info
  - module: community.general.redis
"""

EXAMPLES = r"""
- name: Increment integer key foo on localhost with no username and print new value
  community.general.redis_data_incr:
    login_host: localhost
    login_password: supersecret
    key: foo
    increment_int: 1
  register: result
- name: Print new value
  debug:
    var: result.value

- name: Increment float key foo by 20.4
  community.general.redis_data_incr:
    login_host: redishost
    login_user: redisuser
    login_password: somepass
    key: foo
    increment_float: '20.4'
"""

RETURN = r"""
value:
  description: Incremented value of key.
  returned: on success
  type: float
  sample: '4039.4'
msg:
  description: A short message.
  returned: always
  type: str
  sample: 'Incremented key: foo by 20.4 to 65.9'
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._redis import (
    RedisAnsible,
    fail_imports,
    redis_auth_argument_spec,
)


def main():
    redis_auth_args = redis_auth_argument_spec()
    module_args = dict(
        key=dict(type="str", required=True, no_log=False),
        increment_int=dict(type="int"),
        increment_float=dict(type="float"),
    )
    module_args.update(redis_auth_args)

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True,
        mutually_exclusive=[["increment_int", "increment_float"]],
    )
    fail_imports(module)

    redis = RedisAnsible(module)
    key = module.params["key"]
    increment_float = module.params["increment_float"]
    increment_int = module.params["increment_int"]
    increment = 1
    if increment_float is not None:
        increment = increment_float
    elif increment_int is not None:
        increment = increment_int

    result = {"changed": False}
    if module.check_mode:
        value = 0.0
        try:
            res = redis.connection.get(key)
            if res is not None:
                value = float(res)
        except ValueError:
            msg = f"Value: {res} of key: {key} is not incrementable(int or float)"
            result["msg"] = msg
            module.fail_json(**result)
        except Exception as e:
            msg = f"Failed to get value of key: {key} with exception: {e}"
            result["msg"] = msg
            module.fail_json(**result)
        msg = f"Incremented key: {key} by {increment} to {value + increment}"
        result["msg"] = msg
        result["value"] = float(value + increment)
        module.exit_json(**result)

    if increment_float is not None:
        try:
            value = redis.connection.incrbyfloat(key, increment)
            msg = f"Incremented key: {key} by {increment} to {value}"
            result["msg"] = msg
            result["value"] = float(value)
            result["changed"] = True
            module.exit_json(**result)
        except Exception as e:
            msg = f"Failed to increment key: {key} by {increment} with exception: {e}"
            result["msg"] = msg
            module.fail_json(**result)
    elif increment_int is not None:
        try:
            value = redis.connection.incrby(key, increment)
            msg = f"Incremented key: {key} by {increment} to {value}"
            result["msg"] = msg
            result["value"] = float(value)
            result["changed"] = True
            module.exit_json(**result)
        except Exception as e:
            msg = f"Failed to increment key: {key} by {increment} with exception: {e}"
            result["msg"] = msg
            module.fail_json(**result)
    else:
        try:
            value = redis.connection.incr(key)
            msg = f"Incremented key: {key} to {value}"
            result["msg"] = msg
            result["value"] = float(value)
            result["changed"] = True
            module.exit_json(**result)
        except Exception as e:
            msg = f"Failed to increment key: {key} with exception: {e}"
            result["msg"] = msg
            module.fail_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2018, Juergen Wiebe <wiebe@e-spirit.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: utm_network_interface_address_info

author:
  - Juergen Wiebe (@steamx)

short_description: Get info for a network/interface_address object

description:
  - Get info for a network/interface_address object in SOPHOS UTM.
attributes:
  check_mode:
    version_added: 3.3.0
    # This was backported to 2.5.4 and 1.3.11 as well, since this was a bugfix

options:
  name:
    type: str
    description:
      - The name of the object that identifies the entry.
    required: true

extends_documentation_fragment:
  - community.general._utm
  - community.general._attributes
  - community.general._attributes.info_module
"""

EXAMPLES = r"""
- name: Get network interface address info
  utm_proxy_interface_address_info:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestNetworkInterfaceAddress
"""

RETURN = r"""
result:
  description: The utm object that was created.
  returned: success
  type: complex
  contains:
    _ref:
      description: The reference name of the object.
      type: str
    _locked:
      description: Whether or not the object is currently locked.
      type: bool
    _type:
      description: The type of the object.
      type: str
    name:
      description: The name of the object.
      type: str
    address:
      description: The ip4 address of the network/interface_address object.
      type: str
    address6:
      description: The ip6 address of the network/interface_address object.
      type: str
    comment:
      description: The comment string.
      type: str
    resolved:
      description: Whether or not the object is resolved.
      type: bool
    resolved6:
      description: Whether or not the object is resolved.
      type: bool
"""

from ansible_collections.community.general.plugins.module_utils._utm_utils import UTM, UTMModule


def main():
    endpoint = "network/interface_address"
    key_to_check_for_changes = []
    module = UTMModule(
        argument_spec=dict(name=dict(type="str", required=True)),
        supports_check_mode=True,
    )
    try:
        UTM(module, endpoint, key_to_check_for_changes, info_only=True).execute()
    except Exception as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2012, Matt Wright <matt@nobien.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: supervisorctl
short_description: Manage the state of a program or group of programs managed by C(supervisord)
description:
  - Manage the state of a program or group of programs managed by C(supervisord).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    type: str
    description:
      - The name of the supervisord program or group to manage.
      - The name is taken as group name when it ends with a colon V(:).
      - If O(name=all), all programs and program groups are managed.
    required: true
  config:
    type: path
    description:
      - The supervisor configuration file path.
  server_url:
    type: str
    description:
      - URL on which supervisord server is listening.
  username:
    type: str
    description:
      - Username to use for authentication.
  password:
    type: str
    description:
      - Password to use for authentication.
  state:
    type: str
    description:
      - The desired state of program/group.
    required: true
    choices: ["present", "started", "stopped", "restarted", "absent", "signalled"]
  stop_before_removing:
    type: bool
    description:
      - Use O(stop_before_removing=true) to stop the program/group before removing it.
    default: false
    version_added: 7.5.0
  signal:
    type: str
    description:
      - The signal to send to the program/group, when combined with the V(signalled) state. Required when l(state=signalled).
  supervisorctl_path:
    type: path
    description:
      - Path to C(supervisorctl) executable.
notes:
  - When O(state=present), the module calls C(supervisorctl reread) then C(supervisorctl add) if the program/group does not
    exist.
  - When O(state=restarted), the module calls C(supervisorctl update) then calls C(supervisorctl restart).
  - When O(state=absent), the module calls C(supervisorctl reread) then C(supervisorctl remove) to remove the target program/group.
    If the program/group is still running, the action fails. If you want to stop the program/group before removing, use O(stop_before_removing=true).
requirements: ["supervisorctl"]
author:
  - "Matt Wright (@mattupstate)"
  - "Aaron Wang (@inetfuture) <inetfuture@gmail.com>"
"""

EXAMPLES = r"""
- name: Manage the state of program to be in started state
  community.general.supervisorctl:
    name: my_app
    state: started

- name: Manage the state of program group to be in started state
  community.general.supervisorctl:
    name: 'my_apps:'
    state: started

- name: Restart my_app, reading supervisorctl configuration from a specified file
  community.general.supervisorctl:
    name: my_app
    state: restarted
    config: /var/opt/my_project/supervisord.conf

- name: Restart my_app, connecting to supervisord with credentials and server URL
  community.general.supervisorctl:
    name: my_app
    state: restarted
    username: test
    password: testpass
    server_url: http://localhost:9001

- name: Send a signal to my_app via supervisorctl
  community.general.supervisorctl:
    name: my_app
    state: signalled
    signal: USR1

- name: Restart all programs and program groups
  community.general.supervisorctl:
    name: all
    state: restarted
"""


import os

from ansible.module_utils.basic import AnsibleModule, is_executable


def _is_active(status):
    return status in ("RUNNING", "STARTING")


def _is_not_active(status):
    return not _is_active(status)


def _is_strictly_running(status):
    return status in ("RUNNING",)


def _always(_status):
    return True


def main():
    arg_spec = dict(
        name=dict(type="str", required=True),
        config=dict(type="path"),
        server_url=dict(type="str"),
        username=dict(type="str"),
        password=dict(type="str", no_log=True),
        supervisorctl_path=dict(type="path"),
        state=dict(
            type="str", required=True, choices=["present", "started", "restarted", "stopped", "absent", "signalled"]
        ),
        stop_before_removing=dict(type="bool", default=False),
        signal=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=arg_spec,
        supports_check_mode=True,
        required_if=[("state", "signalled", ["signal"])],
    )

    name = module.params["name"]
    is_group = False
    if name.endswith(":"):
        is_group = True
        name = name.rstrip(":")
    state = module.params["state"]
    stop_before_removing = module.params.get("stop_before_removing")
    config = module.params.get("config")
    server_url = module.params.get("server_url")
    username = module.params.get("username")
    password = module.params.get("password")
    supervisorctl_path = module.params.get("supervisorctl_path")
    signal = module.params.get("signal")

    # we check error message for a pattern, so we need to make sure that's in C locale
    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    if supervisorctl_path:
        if os.path.exists(supervisorctl_path) and is_executable(supervisorctl_path):
            supervisorctl_args = [supervisorctl_path]
        else:
            module.fail_json(
                msg=f"Provided path to supervisorctl does not exist or isn't executable: {supervisorctl_path}"
            )
    else:
        supervisorctl_args = [module.get_bin_path("supervisorctl", True)]

    if config:
        supervisorctl_args.extend(["-c", config])
    if server_url:
        supervisorctl_args.extend(["-s", server_url])
    if username:
        supervisorctl_args.extend(["-u", username])
    if password:
        supervisorctl_args.extend(["-p", password])

    def run_supervisorctl(cmd, name=None, **kwargs):
        args = list(supervisorctl_args)  # copy the master args
        args.append(cmd)
        if name:
            args.append(name)
        return module.run_command(args, **kwargs)

    def get_matched_processes():
        matched = []
        rc, out, err = run_supervisorctl("status")
        for line in out.splitlines():
            # One status line may look like one of these two:
            # process not in group:
            #   echo_date_lonely RUNNING pid 7680, uptime 13:22:18
            # process in group:
            #   echo_date_group:echo_date_00 RUNNING pid 7681, uptime 13:22:18
            fields = [field for field in line.split(" ") if field != ""]
            process_name = fields[0]
            status = fields[1]

            if is_group:
                # If there is ':', this process must be in a group.
                if ":" in process_name:
                    group = process_name.split(":")[0]
                    if group != name:
                        continue
                else:
                    continue
            else:
                if process_name != name and name != "all":
                    continue

            matched.append((process_name, status))
        return matched

    def take_action_on_processes(processes, status_filter, action, expected_result, exit_module=True):
        to_take_action_on = []
        for process_name, status in processes:
            if status_filter(status):
                to_take_action_on.append(process_name)

        if len(to_take_action_on) == 0:
            if not exit_module:
                return
            module.exit_json(changed=False, name=name, state=state)
        if module.check_mode:
            if not exit_module:
                return
            module.exit_json(changed=True)
        for process_name in to_take_action_on:
            rc, out, err = run_supervisorctl(action, process_name, check_rc=True)
            if f"{process_name}: {expected_result}" not in out:
                module.fail_json(msg=out)

        if exit_module:
            module.exit_json(changed=True, name=name, state=state, affected=to_take_action_on)

    if name == "all" and state in ("started", "stopped", "restarted"):
        if state == "restarted":
            run_supervisorctl("update", check_rc=True)
        processes = get_matched_processes()
        status_filters = {
            "started": _is_not_active,
            "stopped": _is_active,
            "restarted": _always,
        }
        to_act_on = [p for p, s in processes if status_filters[state](s)]
        if not to_act_on:
            module.exit_json(changed=False, name=name, state=state)
        if module.check_mode:
            module.exit_json(changed=True)
        actions = {"started": "start", "stopped": "stop", "restarted": "restart"}
        run_supervisorctl(actions[state], "all", check_rc=True)
        module.exit_json(changed=True, name=name, state=state, affected=to_act_on)

    if state == "restarted":
        rc, out, err = run_supervisorctl("update", check_rc=True)
        processes = get_matched_processes()
        if len(processes) == 0:
            module.fail_json(name=name, msg="ERROR (no such process)")

        take_action_on_processes(processes, _always, "restart", "started")

    processes = get_matched_processes()

    if state == "absent":
        if len(processes) == 0:
            module.exit_json(changed=False, name=name, state=state)

        if stop_before_removing:
            take_action_on_processes(processes, _is_active, "stop", "stopped", exit_module=False)

        if module.check_mode:
            module.exit_json(changed=True)
        run_supervisorctl("reread", check_rc=True)
        rc, out, err = run_supervisorctl("remove", name)
        if f"{name}: removed process group" in out:
            module.exit_json(changed=True, name=name, state=state)
        else:
            module.fail_json(msg=out, name=name, state=state)

    if state == "present":
        if len(processes) > 0:
            module.exit_json(changed=False, name=name, state=state)

        if module.check_mode:
            module.exit_json(changed=True)
        run_supervisorctl("reread", check_rc=True)
        dummy, out, dummy = run_supervisorctl("add", name)
        if f"{name}: added process group" in out:
            module.exit_json(changed=True, name=name, state=state)
        else:
            module.fail_json(msg=out, name=name, state=state)

    # from this point onwards, if there are no matching processes, module cannot go on.
    if len(processes) == 0 and name != "all":
        module.fail_json(name=name, msg="ERROR (no such process)")

    if state == "started":
        take_action_on_processes(processes, _is_not_active, "start", "started")

    if state == "stopped":
        take_action_on_processes(processes, _is_active, "stop", "stopped")

    if state == "signalled":
        take_action_on_processes(processes, _is_strictly_running, f"signal {signal}", "signalled")


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016-2017, Hewlett Packard Enterprise Development LP
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: oneview_enclosure_info
short_description: Retrieve information about one or more Enclosures
description:
  - Retrieve information about one or more of the Enclosures from OneView.
requirements:
  - hpOneView >= 2.0.1
author:
  - Felipe Bulsoni (@fgbulsoni)
  - Thiago Miotto (@tmiotto)
  - Adriane Cardozo (@adriane-cardozo)
attributes:
  check_mode:
    version_added: 3.3.0
    # This was backported to 2.5.4 and 1.3.11 as well, since this was a bugfix
options:
  name:
    description:
      - Enclosure name.
    type: str
  options:
    description:
      - 'List with options to gather additional information about an Enclosure and related resources. Options allowed: V(script),
        V(environmentalConfiguration), and V(utilization). For the option V(utilization), you can provide specific parameters.'
    type: list
    elements: raw

extends_documentation_fragment:
  - community.general._oneview
  - community.general._oneview.factsparams
  - community.general._attributes
  - community.general._attributes.info_module
"""

EXAMPLES = r"""
- name: Gather information about all Enclosures
  community.general.oneview_enclosure_info:
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  no_log: true
  delegate_to: localhost
  register: result

- name: Print fetched information about Enclosures
  ansible.builtin.debug:
    msg: "{{ result.enclosures }}"

- name: Gather paginated, filtered and sorted information about Enclosures
  community.general.oneview_enclosure_info:
    params:
      start: 0
      count: 3
      sort: name:descending
      filter: status=OK
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  no_log: true
  delegate_to: localhost
  register: result

- name: Print fetched information about paginated, filtered and sorted list of Enclosures
  ansible.builtin.debug:
    msg: "{{ result.enclosures }}"

- name: Gather information about an Enclosure by name
  community.general.oneview_enclosure_info:
    name: Enclosure-Name
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  no_log: true
  delegate_to: localhost
  register: result

- name: Print fetched information about Enclosure found by name
  ansible.builtin.debug:
    msg: "{{ result.enclosures }}"

- name: Gather information about an Enclosure by name with options
  community.general.oneview_enclosure_info:
    name: Test-Enclosure
    options:
      - script                     # optional
      - environmentalConfiguration # optional
      - utilization                # optional
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  no_log: true
  delegate_to: localhost
  register: result

- name: Print fetched information about Enclosure found by name
  ansible.builtin.debug:
    msg: "{{ result.enclosures }}"

- name: Print fetched information about Enclosure Script
  ansible.builtin.debug:
    msg: "{{ result.enclosure_script }}"

- name: Print fetched information about Enclosure Environmental Configuration
  ansible.builtin.debug:
    msg: "{{ result.enclosure_environmental_configuration }}"

- name: Print fetched information about Enclosure Utilization
  ansible.builtin.debug:
    msg: "{{ result.enclosure_utilization }}"

- name: "Gather information about an Enclosure with temperature data at a resolution of one sample per day, between two
    specified dates"
  community.general.oneview_enclosure_info:
    name: Test-Enclosure
    options:
      - utilization:               # optional
          fields: AmbientTemperature
          filter:
            - startDate=2016-07-01T14:29:42.000Z
            - endDate=2017-07-01T03:29:42.000Z
          view: day
          refresh: false
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  no_log: true
  delegate_to: localhost
  register: result

- name: Print fetched information about Enclosure found by name
  ansible.builtin.debug:
    msg: "{{ result.enclosures }}"

- name: Print fetched information about Enclosure Utilization
  ansible.builtin.debug:
    msg: "{{ result.enclosure_utilization }}"
"""

RETURN = r"""
enclosures:
  description: Has all the OneView information about the Enclosures.
  returned: Always, but can be null.
  type: dict

enclosure_script:
  description: Has all the OneView information about the script of an Enclosure.
  returned: When requested, but can be null.
  type: str

enclosure_environmental_configuration:
  description: Has all the OneView information about the environmental configuration of an Enclosure.
  returned: When requested, but can be null.
  type: dict

enclosure_utilization:
  description: Has all the OneView information about the utilization of an Enclosure.
  returned: When requested, but can be null.
  type: dict
"""

from ansible_collections.community.general.plugins.module_utils._oneview import OneViewModuleBase


class EnclosureInfoModule(OneViewModuleBase):
    argument_spec = dict(name=dict(type="str"), options=dict(type="list", elements="raw"), params=dict(type="dict"))

    def __init__(self):
        super().__init__(
            additional_arg_spec=self.argument_spec,
            supports_check_mode=True,
        )

    def execute_module(self):
        info = {}

        if self.module.params["name"]:
            enclosures = self._get_by_name(self.module.params["name"])

            if self.options and enclosures:
                info = self._gather_optional_info(self.options, enclosures[0])
        else:
            enclosures = self.oneview_client.enclosures.get_all(**self.facts_params)

        info["enclosures"] = enclosures

        return dict(changed=False, **info)

    def _gather_optional_info(self, options, enclosure):
        enclosure_client = self.oneview_client.enclosures
        info = {}

        if options.get("script"):
            info["enclosure_script"] = enclosure_client.get_script(enclosure["uri"])
        if options.get("environmentalConfiguration"):
            env_config = enclosure_client.get_environmental_configuration(enclosure["uri"])
            info["enclosure_environmental_configuration"] = env_config
        if options.get("utilization"):
            info["enclosure_utilization"] = self._get_utilization(enclosure, options["utilization"])

        return info

    def _get_utilization(self, enclosure, params):
        fields = view = refresh = filter = ""

        if isinstance(params, dict):
            fields = params.get("fields")
            view = params.get("view")
            refresh = params.get("refresh")
            filter = params.get("filter")

        return self.oneview_client.enclosures.get_utilization(
            enclosure["uri"], fields=fields, filter=filter, refresh=refresh, view=view
        )

    def _get_by_name(self, name):
        return self.oneview_client.enclosures.get_by("name", name)


def main():
    EnclosureInfoModule().run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2025, Marcos Alano <marcoshalano@gmail.com>
# Based on gio_mime module. Copyright (c) 2022, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# In memory: This code is dedicated to my late grandmother, Maria Marlene. 1936-2025. Rest in peace, grandma.
# -Marcos Alano-

# TODO: Add support for diff mode

from __future__ import annotations

DOCUMENTATION = r"""
module: xdg_mime
author:
  - "Marcos Alano (@mhalano)"
short_description: Set default handler for MIME types, for applications using XDG tools
version_added: 10.7.0
description:
  - This module allows configuring the default handler for specific MIME types when you use applications that rely on XDG.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  mime_types:
    description:
      - One or more MIME types for which a default handler is set.
    type: list
    elements: str
    required: true
  handler:
    description:
      - Sets the default handler for the specified MIME types.
      - The desktop file must be installed in the system. If the desktop file is not installed, the module does not fail,
        but the handler is not set either.
      - You must pass a handler in the form V(*.desktop), otherwise the module fails.
    type: str
    required: true
notes:
  - This module is a thin wrapper around C(xdg-mime) tool.
  - See man xdg-mime(1) for more details.
seealso:
  - name: C(xdg-mime) command manual page
    description: Manual page for the command.
    link: https://portland.freedesktop.org/doc/xdg-mime.html
  - name: xdg-utils Documentation
    description: Reference documentation for xdg-utils.
    link: https://www.freedesktop.org/wiki/Software/xdg-utils/
"""

EXAMPLES = r"""
- name: Set Chrome as the default handler for HTTPS
  community.general.xdg_mime:
    mime_types: x-scheme-handler/https
    handler: google-chrome.desktop
  register: result

- name: Set Chrome as the default handler for both HTTP and HTTPS
  community.general.xdg_mime:
    mime_types:
      - x-scheme-handler/http
      - x-scheme-handler/https
    handler: google-chrome.desktop
  register: result
"""

RETURN = r"""
current_handlers:
  description:
    - Currently set handlers for the passed MIME types.
  returned: success
  type: list
  elements: str
  sample:
    - google-chrome.desktop
    - firefox.desktop
version:
  description: Version of the C(xdg-mime) tool.
  type: str
  returned: always
  sample: "1.2.1"
"""

from ansible_collections.community.general.plugins.module_utils._module_helper import ModuleHelper
from ansible_collections.community.general.plugins.module_utils._xdg_mime import xdg_mime_get, xdg_mime_runner


class XdgMime(ModuleHelper):
    output_params = ["handler"]

    module = dict(
        argument_spec=dict(
            mime_types=dict(type="list", elements="str", required=True),
            handler=dict(type="str", required=True),
        ),
        supports_check_mode=True,
    )

    def __init_module__(self):
        self.runner = xdg_mime_runner(self.module, check_rc=True)

        with self.runner("version") as ctx:
            rc, out, err = ctx.run()
            self.vars.version = out.replace("xdg-mime ", "").strip()

        if not self.vars.handler.endswith(".desktop"):
            self.do_raise(msg="Handler must be a .desktop file")

        self.vars.current_handlers = []
        for mime in self.vars.mime_types:
            handler_value = xdg_mime_get(self.runner, mime)
            if not handler_value:
                handler_value = ""
            self.vars.current_handlers.append(handler_value)

    def __run__(self):
        check_mode_return = (0, "Module executed in check mode", "")

        if any(h != self.vars.handler for h in self.vars.current_handlers):
            self.changed = True

        if self.has_changed():
            with self.runner.context(
                args_order="default handler mime_types", check_mode_skip=True, check_mode_return=check_mode_return
            ) as ctx:
                rc, out, err = ctx.run()
                self.vars.stdout = out
                self.vars.stderr = err
                self.vars.set("run_info", ctx.run_info, verbosity=1)


def main():
    XdgMime.execute()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2024, Lincoln Wallace (locnnil) <lincoln.wallace@canonical.com>
# Copyright (c) 2021, Alexei Znamensky (russoz) <russoz@gmail.com>
# Copyright (c) 2021, Marcus Rickert <marcus.rickert@web.de>
# Copyright (c) 2018, Stanislas Lange (angristan) <angristan@pm.me>
# Copyright (c) 2018, Victor Carceler <vcarceler@iespuigcastellar.xeill.net>

# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: snap
short_description: Manages snaps
description:
  - Manages snaps packages.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of the snaps to be installed.
      - Any named snap accepted by the C(snap) command is valid.
      - O(dangerous=true) may be necessary when installing C(.snap) files. See O(dangerous) for more details.
      - The special name V(system) refers to the snapd system-wide configuration namespace.
        When used with O(options), it runs C(snap set system <key=value>).
        It is always considered present and only V(state=present) and O(options) are meaningful for it.
        Support for V(system) was added in community.general 13.0.0.
    required: true
    type: list
    elements: str
  state:
    description:
      - Desired state of the package.
      - When O(state=present) the module uses C(snap install) if the snap is not installed, and C(snap refresh) if it is installed
        but from a different channel.
    default: present
    choices: [absent, present, enabled, disabled]
    type: str
  classic:
    description:
      - Install a snap that has classic confinement.
      - This option corresponds to the C(--classic) argument of the C(snap install) command.
      - This level of confinement is permissive, granting full system access, similar to that of traditionally packaged applications
        that do not use sandboxing mechanisms. This option can only be specified when the task involves a single snap.
      - See U(https://snapcraft.io/docs/snap-confinement) for more details about classic confinement and confinement levels.
    type: bool
    default: false
  channel:
    description:
      - Define which release of a snap is installed and tracked for updates. This option can only be specified if there is
        a single snap in the task.
      - If not passed, the C(snap) command defaults to V(stable).
      - If the value passed does not contain the C(track), it defaults to C(latest). For example, if V(edge) is passed, the
        module assumes the channel to be V(latest/edge).
      - See U(https://snapcraft.io/docs/channels) for more details about snap channels.
    type: str
  options:
    description:
      - Set options with pattern C(key=value) or C(snap:key=value). If a snap name is given, the option is applied to that
        snap only. If the snap name is omitted, the options are applied to all snaps listed in O(name). Options are only applied
        to active snaps.
      - Options are only applied when C(state) is set to V(present). This is done after the necessary installation or refresh
        (upgrade/downgrade) of all the snaps listed in O(name).
      - See U(https://snapcraft.io/docs/configuration-in-snaps) for more details about snap configuration options.
    type: list
    elements: str
    version_added: 4.4.0
  dangerous:
    description:
      - Install the snap in dangerous mode, without validating its assertions and signatures.
      - This is useful when installing local snaps that are either unsigned or have signatures that have not been acknowledged.
      - See U(https://snapcraft.io/docs/install-modes) for more details about installation modes.
    type: bool
    default: false
    version_added: 7.2.0
  devmode:
    description:
      - Install the snap in developer mode, granting the snap full system access and disabling security confinement.
      - See U(https://snapcraft.io/docs/install-modes) for more details about installation modes.
    type: bool
    default: false
    version_added: 13.0.0
  revision:
    description:
      - Install a specific revision of the snap.
      - This option can only be specified if there is a single snap in the task.
      - Mutually exclusive with O(channel).
      - When a specific revision is set, the snap is held (C(snap refresh --hold)) to prevent automatic updates from
        overriding the pinned revision.
      - "B(Note:) running C(snap refresh <name>) manually bypasses the hold and will update the snap regardless.
        The hold only prevents snapd's automatic background refreshes.
        See U(https://snapcraft.io/docs/how-to-guides/manage-snaps/manage-updates/) for details."
      - See U(https://snapcraft.io/docs/revisions) for more details about snap revisions.
    type: int
    version_added: 13.0.0
notes:
  - Privileged operations, such as installing and configuring snaps, require root priviledges. This is only the case if the
    user has not logged in to the Snap Store.
author:
  - Victor Carceler (@vcarceler) <vcarceler@iespuigcastellar.xeill.net>
  - Stanislas Lange (@angristan) <angristan@pm.me>

seealso:
  - module: community.general.snap_alias
"""

EXAMPLES = r"""
# Install "foo" and "bar" snap
- name: Install foo
  community.general.snap:
    name:
      - foo
      - bar

# Install "foo" snap with options par1=A and par2=B
- name: Install "foo" with options
  community.general.snap:
    name:
      - foo
    options:
      - par1=A
      - par2=B

# Install "foo" and "bar" snaps with common option com=A and specific options fooPar=X and barPar=Y
- name: Install "foo" and "bar" with options
  community.general.snap:
    name:
      - foo
      - bar
    options:
      - com=A
      - foo:fooPar=X
      - bar:barPar=Y

# Remove "foo" snap
- name: Remove foo
  community.general.snap:
    name: foo
    state: absent

# Install a snap with classic confinement
- name: Install "foo" with option --classic
  community.general.snap:
    name: foo
    classic: true

# Install a snap with from a specific channel
- name: Install "foo" with option --channel=latest/edge
  community.general.snap:
    name: foo
    channel: latest/edge

# Install a specific revision of a snap (automatically held to prevent auto-updates)
- name: Install revision 481 of "helm"
  community.general.snap:
    name: helm
    classic: true
    revision: 481

# Set snap system-wide configuration
- name: Configure snapd proxy
  community.general.snap:
    name: system
    options:
      - proxy.http=http://proxy.example.com:3128/
      - proxy.https=http://proxy.example.com:3128/
"""

RETURN = r"""
classic:
  description: Whether or not the snaps were installed with the classic confinement.
  type: bool
  returned: When snaps are installed
channel:
  description: The channel the snaps were installed from.
  type: str
  returned: When snaps are installed
cmd:
  description: The command that was executed on the host.
  type: str
  returned: When changed is true
snaps_installed:
  description: The list of actually installed snaps.
  type: list
  returned: When any snaps have been installed
snaps_removed:
  description: The list of actually removed snaps.
  type: list
  returned: When any snaps have been removed
options_changed:
  description: The list of options set/changed in format C(snap:key=value).
  type: list
  returned: When any options have been changed/set
  version_added: 4.4.0
revision:
  description: The revision of the snap that was installed.
  type: int
  returned: When snaps are installed with a specific revision
  version_added: 13.0.0
version:
  description: Versions of snap components as reported by C(snap version).
  type: dict
  returned: always
  version_added: 10.3.0
"""

import json
import numbers
import re

from ansible.module_utils.common.text.converters import to_native

_VIRTUAL_SNAPS = frozenset({"system"})

from ansible_collections.community.general.plugins.module_utils._module_helper import StateModuleHelper
from ansible_collections.community.general.plugins.module_utils._snap import get_version, snap_runner


class Snap(StateModuleHelper):
    NOT_INSTALLED = 0
    CHANNEL_MISMATCH = 1
    INSTALLED = 2
    REVISION_MISMATCH = 3
    HOLD_MISMATCH = 4

    __disable_re = re.compile(r"(?:\S+\s+){5}(?P<notes>\S+)")
    __set_param_re = re.compile(r"(?P<snap_prefix>\S+:)?(?P<key>\S+)\s*=\s*(?P<value>.+)")
    __list_re = re.compile(r"^(?P<name>\S+)\s+\S+\s+(?P<rev>\S+)\s+(?P<channel>\S+)\s+\S+\s+(?P<notes>\S+)")
    module = dict(
        argument_spec={
            "name": dict(type="list", elements="str", required=True),
            "state": dict(type="str", default="present", choices=["absent", "present", "enabled", "disabled"]),
            "classic": dict(type="bool", default=False),
            "channel": dict(type="str"),
            "options": dict(type="list", elements="str"),
            "dangerous": dict(type="bool", default=False),
            "devmode": dict(type="bool", default=False),
            "revision": dict(type="int"),
        },
        mutually_exclusive=[["channel", "revision"]],
        supports_check_mode=True,
    )

    @staticmethod
    def _first_non_zero(a):
        for elem in a:
            if elem != 0:
                return elem

        return 0

    def __init_module__(self):
        self.runner = snap_runner(self.module)
        self.vars.version = get_version(self.runner)
        # if state=present there might be file names passed in 'name', in
        # which case they must be converted to their actual snap names, which
        # is done using the names_from_snaps() method calling 'snap info'.
        self.vars.set("snapinfo_run_info", [], output=(self.verbosity >= 4))
        self.vars.set("status_run_info", [], output=(self.verbosity >= 4))
        self.vars.set("status_out", None, output=(self.verbosity >= 4))
        self.vars.set("run_info", [], output=(self.verbosity >= 4))

        if self.vars.state == "present":
            self.vars.set("snap_names", self.names_from_snaps(self.vars.name))
            status_var = "snap_names"
        else:
            status_var = "name"
        self.vars.set("status_var", status_var, output=False)
        self.vars.set(
            "snap_status",
            self.snap_status(self.vars[self.vars.status_var], self.vars.channel, self.vars.revision),
            output=False,
            change=True,
        )
        self.vars.set("snap_status_map", dict(zip(self.vars.name, self.vars.snap_status)), output=False, change=True)

    def __quit_module__(self):
        self.vars.snap_status = self.snap_status(self.vars[self.vars.status_var], self.vars.channel, self.vars.revision)
        if self.vars.channel is None:
            self.vars.channel = "stable"

    def _run_multiple_commands(self, commands, actionable_names, bundle=True, refresh=False):
        results_cmd = []
        results_rc = []
        results_out = []
        results_err = []
        results_run_info = []

        state = "refresh" if refresh else self.vars.state

        with self.runner(commands + ["name"]) as ctx:
            if bundle:
                rc, out, err = ctx.run(state=state, name=actionable_names)
                results_cmd.append(commands + actionable_names)
                results_rc.append(rc)
                results_out.append(out.strip())
                results_err.append(err.strip())
                results_run_info.append(ctx.run_info)
            else:
                for name in actionable_names:
                    rc, out, err = ctx.run(state=state, name=name)
                    results_cmd.append(commands + [name])
                    results_rc.append(rc)
                    results_out.append(out.strip())
                    results_err.append(err.strip())
                    results_run_info.append(ctx.run_info)

        return (
            "; ".join([to_native(x) for x in results_cmd]),
            self._first_non_zero(results_rc),
            "\n".join(results_out),
            "\n".join(results_err),
            results_run_info,
        )

    def convert_json_subtree_to_map(self, json_subtree, prefix=None):
        option_map = {}

        if not isinstance(json_subtree, dict):
            self.do_raise(
                "Non-dict non-leaf element encountered while parsing option map. "
                "The output format of 'snap set' may have changed. Aborting!"
            )

        for key, value in json_subtree.items():
            full_key = key if prefix is None else f"{prefix}.{key}"

            if isinstance(value, (str, float, bool, numbers.Integral)):
                option_map[full_key] = str(value)
            else:
                option_map.update(self.convert_json_subtree_to_map(json_subtree=value, prefix=full_key))

        return option_map

    def convert_json_to_map(self, json_string):
        json_object = json.loads(json_string)
        return self.convert_json_subtree_to_map(json_object)

    def retrieve_option_map(self, snap_name):
        with self.runner("get name") as ctx:
            rc, out, err = ctx.run(name=snap_name)

        if rc != 0:
            return {}

        result = out.splitlines()

        if "has no configuration" in result[0]:
            return {}

        try:
            option_map = self.convert_json_to_map(out)
            return option_map
        except Exception as e:
            self.do_raise(
                msg=f"Parsing option map returned by 'snap get {snap_name}' triggers exception '{e}', output:\n'{out}'"
            )

    def names_from_snaps(self, snaps):
        real_snaps = [s for s in snaps if s not in _VIRTUAL_SNAPS]

        if not real_snaps:
            return list(snaps)

        def process_one(rc, out, err):
            res = [line for line in out.split("\n") if line.startswith("name:")]
            name = res[0].split()[1]
            return [name]

        def process_many(rc, out, err):
            # This needs to be "\n---" instead of just "---" because otherwise
            # if a snap uses "---" in its description then that will incorrectly
            # be interpreted as a separator between snaps in the output.
            outputs = out.split("\n---")
            res = []
            for sout in outputs:
                res.extend(process_one(rc, sout, ""))
            return res

        def process(rc, out, err):
            if len(real_snaps) == 1:
                check_error = err
                process_ = process_one
            else:
                check_error = out
                process_ = process_many

            if "warning: no snap found" in check_error:
                snaps_not_found = [x.split()[-1] for x in out.split("\n") if x.startswith("warning: no snap found")]
                self.do_raise(f"Snaps not found: {snaps_not_found}.")
            return process_(rc, out, err)

        real_names = []
        with self.runner("info name", output_process=process) as ctx:
            try:
                real_names = ctx.run(name=real_snaps)
            finally:
                self.vars.snapinfo_run_info.append(ctx.run_info)

        real_name_iter = iter(real_names)
        return [s if s in _VIRTUAL_SNAPS else next(real_name_iter) for s in snaps]

    def snap_status(self, snap_name, channel, revision=None):
        should_be_held = revision is not None

        def _status_check(name, channel, revision, installed):
            if name in _VIRTUAL_SNAPS:
                return Snap.INSTALLED
            match = [(r, c, notes) for n, r, c, notes in installed if n == name]
            if not match:
                return Snap.NOT_INSTALLED
            installed_rev, installed_channel, installed_notes = match[0]
            if revision is not None and str(revision) != installed_rev:
                return Snap.REVISION_MISMATCH
            if channel and installed_channel not in (channel, f"latest/{channel}"):
                return Snap.CHANNEL_MISMATCH
            if should_be_held:
                is_held = "held" in installed_notes.split(",")
                if not is_held:
                    return Snap.HOLD_MISMATCH
            return Snap.INSTALLED

        with self.runner("_list") as ctx:
            rc, out, err = ctx.run(check_rc=True)
        list_out = out.split("\n")[1:]
        list_out = [self.__list_re.match(x) for x in list_out]
        list_out = [(m.group("name"), m.group("rev"), m.group("channel"), m.group("notes")) for m in list_out if m]
        self.vars.status_out = list_out
        self.vars.status_run_info = ctx.run_info

        return [_status_check(n, channel, revision, list_out) for n in snap_name]

    def is_snap_enabled(self, snap_name):
        with self.runner("_list name") as ctx:
            rc, out, err = ctx.run(name=snap_name)
        if rc != 0:
            return None
        result = out.splitlines()[1]
        match = self.__disable_re.match(result)
        if not match:
            self.do_raise(msg=f"Unable to parse 'snap list {snap_name}' output:\n{out}")
        notes = match.group("notes")
        return "disabled" not in notes.split(",")

    def _present(self, actionable_snaps, refresh=False):
        self.changed = True
        self.vars.snaps_installed = actionable_snaps

        if self.check_mode:
            return

        params = ["state", "classic", "channel", "revision", "dangerous", "devmode"]  # get base cmd parts
        has_one_pkg_params = bool(self.vars.classic) or self.vars.channel != "stable" or self.vars.revision is not None
        has_multiple_snaps = len(actionable_snaps) > 1

        if has_one_pkg_params and has_multiple_snaps:
            self.vars.cmd, rc, out, err, run_info = self._run_multiple_commands(
                params, actionable_snaps, bundle=False, refresh=refresh
            )
        else:
            self.vars.cmd, rc, out, err, run_info = self._run_multiple_commands(
                params, actionable_snaps, refresh=refresh
            )
        self.vars.run_info = run_info

        if rc == 0:
            return

        classic_snap_pattern = re.compile(
            r'^error: This revision of snap "(?P<package_name>\w+)"'
            r" was published using classic confinement"
        )
        match = classic_snap_pattern.match(err)
        if match:
            err_pkg = match.group("package_name")
            msg = f"Couldn't install {err_pkg} because it requires classic confinement"
        else:
            msg = f"Ooops! Snap installation failed while executing '{self.vars.cmd}', please examine logs and error output for more details."
        self.do_raise(msg=msg)

    def _apply_hold(self, snaps):
        if not snaps:
            return
        self.changed = True
        if self.check_mode:
            return
        for snap_name in snaps:
            with self.runner("state hold name") as ctx:
                rc, out, err = ctx.run(state="refresh", hold=True, name=snap_name)
            if rc != 0:
                self.do_raise(msg=f"Snap hold failed for '{snap_name}': {err}")

    def state_present(self):
        self.vars.set_meta("classic", output=True)
        self.vars.set_meta("channel", output=True)
        self.vars.set_meta("revision", output=True)

        actionable_refresh = [
            snap
            for snap in self.vars.name
            if self.vars.snap_status_map[snap] in (Snap.CHANNEL_MISMATCH, Snap.REVISION_MISMATCH)
        ]
        if actionable_refresh:
            self._present(actionable_refresh, refresh=True)
        actionable_install = [snap for snap in self.vars.name if self.vars.snap_status_map[snap] == Snap.NOT_INSTALLED]
        if actionable_install:
            self._present(actionable_install)

        if self.vars.revision is not None:
            hold_mismatch = [snap for snap in self.vars.name if self.vars.snap_status_map[snap] == Snap.HOLD_MISMATCH]
            self._apply_hold(actionable_install + actionable_refresh + hold_mismatch)

        self.set_options()

    def set_options(self):
        if self.vars.options is None:
            return

        actionable_snaps = [s for s in self.vars.name if self.vars.snap_status_map[s] != Snap.NOT_INSTALLED]
        overall_options_changed = []

        for snap_name in actionable_snaps:
            option_map = self.retrieve_option_map(snap_name=snap_name)

            options_changed = []

            for option_string in self.vars.options:
                match = self.__set_param_re.match(option_string)

                if not match:
                    msg = f"Cannot parse set option '{option_string}'"
                    self.do_raise(msg)

                snap_prefix = match.group("snap_prefix")
                selected_snap_name = snap_prefix[:-1] if snap_prefix else None

                if selected_snap_name is not None and selected_snap_name not in self.vars.name:
                    msg = f"Snap option '{option_string}' refers to snap which is not in the list of snap names"
                    self.do_raise(msg)

                if selected_snap_name is None or (snap_name is not None and snap_name == selected_snap_name):
                    key = match.group("key")
                    value = match.group("value").strip()

                    if key not in option_map or key in option_map and option_map[key] != value:
                        option_without_prefix = f"{key}={value}"
                        option_with_prefix = (
                            option_string if selected_snap_name is not None else f"{snap_name}:{option_string}"
                        )
                        options_changed.append(option_without_prefix)
                        overall_options_changed.append(option_with_prefix)

            if options_changed:
                self.changed = True

                if not self.check_mode:
                    with self.runner("_set name options") as ctx:
                        rc, out, err = ctx.run(name=snap_name, options=options_changed)
                    if rc != 0:
                        if 'has no "configure" hook' in err:
                            msg = f"Snap '{snap_name}' does not have any configurable options"
                            self.do_raise(msg)

                        msg = f"Cannot set options '{' '.join(options_changed)}' for snap '{snap_name}': error={err}"
                        self.do_raise(msg)

        if overall_options_changed:
            self.vars.options_changed = overall_options_changed

    def _generic_state_action(self, actionable_func, actionable_var, params):
        actionable_snaps = [s for s in self.vars.name if actionable_func(s)]
        if not actionable_snaps:
            return
        self.changed = True
        self.vars[actionable_var] = actionable_snaps
        if self.check_mode:
            return
        self.vars.cmd, rc, out, err, run_info = self._run_multiple_commands(params, actionable_snaps)
        self.vars.run_info = run_info
        if rc == 0:
            return
        msg = f"Ooops! Snap operation failed while executing '{self.vars.cmd}', please examine logs and error output for more details."
        self.do_raise(msg=msg)

    def state_absent(self):
        self._generic_state_action(
            lambda s: self.vars.snap_status_map[s] != Snap.NOT_INSTALLED,
            "snaps_removed",
            ["classic", "channel", "state"],
        )

    def state_enabled(self):
        self._generic_state_action(
            lambda s: not self.is_snap_enabled(s), "snaps_enabled", ["classic", "channel", "state"]
        )

    def state_disabled(self):
        self._generic_state_action(self.is_snap_enabled, "snaps_disabled", ["classic", "channel", "state"])


def main():
    Snap.execute()


if __name__ == "__main__":
    main()
